"""
Multi-process Parallel Simulation Tool
"""

import queue
import signal
import time, os, shutil, sys
import multiprocessing as mp
from concurrent.futures import CancelledError, Future, ProcessPoolExecutor
from pathlib import Path
from typing import Optional, List, Tuple
from feasytools import time2str, ArgChecker
from v2simux import (
    Lang, PluginPool, StaPool,
    get_sim_params,
    simulate_multi,
    load_external_components,
    simulate_single,
)


EXT_COMP = (Path(__file__).parent / "external_components")
            

def _create_argchk(cmd: str) -> ArgChecker:
    return ArgChecker(cmd, ["plot", "gen-veh", "gen-fcs", "gen-scs"])


class SimCommand:
    """A class to represent a simulation command."""
    def __init__(self, cmd: str, fromfile: bool = True, data_dir: str = ""):
        """
        Create a new simulation command.
            cmd: Command string.
            fromfile: Whether the command is from a file.
            data_dir: Data directory.
        """
        self.args = _create_argchk(cmd)
        self.fromfile = fromfile
        if fromfile:
            self.no_parallel = self.args.pop_bool("no-parallel")
            self.output = self.args.pop_str("o", "")
        else:
            self.args.pop_int("seed", 0)
            self.args.pop_str("o", "")
            self.gen_veh = ArgChecker(self.args.pop_str("gen-veh", ""))
            self.gen_veh.pop_int("seed", 0)
            self.gen_fcs = ArgChecker(self.args.pop_str("gen-fcs", ""))
            self.gen_fcs.pop_int("seed", 0)
            self.gen_scs = ArgChecker(self.args.pop_str("gen-scs", ""))
            self.gen_scs.pop_int("seed", 0)
            self.plot = ArgChecker(self.args.pop_str("plot", ""))
            self.no_parallel = False
            self.data_dir = data_dir

    def __arg2str(self, arg: ArgChecker, seed: Optional[int] = None) -> str:
        ret = ""
        for k, v in arg.items():
            ret += f"-{k} '{v}' "
        if seed is None:
            return ret
        return ret + f"-seed {seed}"

    def get_new_command(self, seed: int, resdir: str) -> str:
        """
        Get the new command string with the given seed and result directory.
            seed: Seed to use.
            resdir: Result directory.
        """
        if self.fromfile:
            return self.__arg2str(self.args) + f" -o '{resdir}'"
        return (
            self.__arg2str(self.args, seed)
            + f" -d '{self.data_dir}' -o '{resdir}'"
            + (
                f' -gen-veh "{self.__arg2str(self.gen_veh, seed)}"'
                if not self.gen_veh.empty()
                else ""
            )
            + (
                f' -gen-fcs "{self.__arg2str(self.gen_fcs, seed)}"'
                if not self.gen_fcs.empty()
                else ""
            )
            + (
                f' -gen-scs "{self.__arg2str(self.gen_scs, seed)}"'
                if not self.gen_scs.empty()
                else ""
            )
            + f' -plot "{self.__arg2str(self.plot)}"'
        )


def work(
    i: int,
    resdir: str,
    cmd: SimCommand,
    plg_pool: PluginPool,
    sta_pool: StaPool,
    mpQ: Optional[queue.Queue] = None,
) -> bool:
    """
    Run a simulation task.
        i: Task index.
        resdir: Result directory.
        cmd: Simulation command.
        plg_pool: Plugin pool.
        sta_pool: Statistics pool.
        mpQ: The multiprocessing queue for communicating with the main process.
    """
    # Create a copy of the simulation folder
    pname = f"_sim_temp/{i:03}"
    Path(pname).mkdir(parents=True, exist_ok=True)
    if "d" in cmd.args:
        data_root = cmd.args["d"]
    else:
        data_root = cmd.data_dir
    shutil.copytree(data_root, pname, dirs_exist_ok=True)

    if cmd.fromfile:
        resdir = str(Path(resdir).parent / cmd.output)
    cmd_text = cmd.get_new_command(i, resdir)
    pars = get_sim_params(_create_argchk(cmd_text), plg_pool, sta_pool)
    pars["cfgdir"] = pname
    if mpQ:
        ret = simulate_multi(mpQ, i, **pars)
    else:
        ret = simulate_single(**pars)

    # Clear the temporary folder
    shutil.rmtree(pname)
    return ret


def parallel_sim(parallel: int, results_root: str, commands: List[SimCommand]):
    """
    Run parallel simulations.
        parallel: Maximum number of parallel tasks.
        results_root: Root directory of the results.
        commands: Simulation commands.
    """
    plg_pool = PluginPool()
    sta_pool = StaPool()
    if EXT_COMP.exists():
        load_external_components(EXT_COMP, plg_pool, sta_pool)
    pool = ProcessPoolExecutor(parallel)
    mpQ:queue.Queue = mp.Manager().Queue()

    skip_list: List[int] = []
    work_list: List[Future] = []
    work_id_list: List[int] = []
    non_paras: List[Tuple[int, SimCommand]] = []
    pre_count = 0

    st_time = time.time()
    for i, cfg in enumerate(commands):
        resdir = Path(results_root + f"/{i:03}")
        if resdir.exists():
            skip_list.append(i)
            continue
        if cfg.no_parallel:
            non_paras.append((i, cfg))
        else:
            work_list.append(
                pool.submit(work, i, str(resdir), cfg, plg_pool, sta_pool, mpQ)
            )
            work_id_list.append(i)

    print(Lang.PARA_SIM_SKIP_LIST.format(skip_list))
    rounds = len(commands)
    progs: List[float] = [0.0] * rounds
    last_upd = 0
    real_rounds = rounds - len(skip_list) - len(non_paras)

    StopSignal = False

    def all_done():
        for f in work_list:
            if not f.done():
                return False
        return True

    def eh(signum, frame):
        nonlocal StopSignal
        print()
        print(Lang.MAIN_SIGINT)
        StopSignal = True
        pool.shutdown()

    signal.signal(signal.SIGINT, eh)

    while not all_done():
        try:
            t = mpQ.get(timeout=1)
            clnt = t.clntID
            con = t.cmd
            obj, sta = con.split(":")
            if obj == "sim":
                if sta == "done":
                    progs[clnt] = 100
                elif sta == "start":
                    progs[clnt] = 0
                    pre_count += 1
                else:
                    progs[clnt] = float(sta)
        except queue.Empty:
            pass
        if time.time() - last_upd > 1 and not StopSignal:
            pre_info = (
                f"Pre: {pre_count}/{real_rounds} " if pre_count < real_rounds else ""
            )
            last_upd = time.time()
            elapsed = time2str(last_upd - st_time)
            avg_prog = sum(progs) / real_rounds
            eta = (
                time2str((last_upd - st_time) * (100 - avg_prog) / avg_prog)
                if avg_prog > 0
                else "N/A"
            )
            print(
                Lang.PARA_SIM_PROG.format(avg_prog, pre_info, elapsed, eta),
                end="\r",
            )
        time.sleep(0.5)

    print()
    shutil.rmtree("_sim_temp", ignore_errors=True)

    for i, f in zip(work_id_list, work_list):
        e = f.exception()
        if e:
            if isinstance(e, (KeyboardInterrupt, CancelledError)):
                continue
            print("Error:", i, type(e).__name__, e.with_traceback(e.__traceback__))

    if len(non_paras) > 0:
        print(Lang.PARA_SIM_DONE_PARA.format(time2str(time.time() - st_time)))
        st_time2 = time.time()
        print(Lang.PARA_SIM_START_SERIAL)
        for i, cfg in non_paras:
            if not work(i, results_root + f"/{i:03}", cfg, plg_pool, sta_pool):
                break  # Stop if any error occurs
        print(Lang.PARA_SIM_DONE_SERIAL.format(time2str(time.time() - st_time2)))
    print(Lang.MAIN_SIM_DONE.format(time2str(time.time() - st_time)))


def main():
    args = ArgChecker(force_parametric=["p","d","r","c","n","f"])
    if args.pop_bool("h") or args.pop_bool("help"):
        print(Lang.PARA_HELP_STR.format(sys.argv[0]))
        exit()

    max_parallel = os.cpu_count()
    if max_parallel is None:
        max_parallel = 1
    else:
        max_parallel -= 1
    parallel = args.pop_int("p", max_parallel)
    root = args.pop_str("r", "results_set")
    from_file = args.pop_str("f", "")
    if from_file == "":
        data_root = args.pop_str("d")
        cmd = args.pop_str("c")
        n = args.pop_int("n", 50)
        commands = [SimCommand(cmd,False,data_root) for _ in range(n)]
    else:
        with open(from_file, "r") as f:
            lines = f.readlines()
        commands = []
        for line in lines:
            comment_start = line.find("#")
            if comment_start != -1:
                line = line[:comment_start]
            line = line.strip()
            if line == "":
                continue
            commands.append(SimCommand(line))
    parallel_sim(parallel, root, commands)


if __name__ == "__main__":
    main()


from v2simux_gui.common import *

from .optbox import OptionBox
from .plotpad import PlotPad


_L = LangLib.Load(__file__)
AVAILABLE_ITEMS = ["fcs","scs","ev","gen","bus","line","pvw","ess"]
AVAILABLE_ITEMS2 = AVAILABLE_ITEMS + ["fcs_accum","scs_accum","bus_total","gen_total"]


class PlotPage(Frame):
    @property
    def AccumPlotMax(self)->bool:
        return self.accum_plotmax.get()
    
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.columnconfigure(index=0,weight=1)
        self.columnconfigure(index=1,weight=1)
        self.lfra_head = LabelFrame(self, text=_L["TIME"])
        self.lfra_head.grid(row=0,column=0,padx=3,pady=5, columnspan=2, sticky="nsew")
        self.panel_time = Frame(self.lfra_head)
        self.panel_time.pack(side="top",fill="x",anchor='w',pady=2,) 
        self.lb_time = Label(self.panel_time, text=_L["START_TIME"])
        self.lb_time.pack(side="left")
        self.entry_time = Entry(self.panel_time,width=10)
        self.entry_time.insert(0,"86400")
        self.entry_time.pack(side="left")
        self.lb_end_time = Label(self.panel_time, text=_L["END_TIME"])
        self.lb_end_time.pack(side="left")
        self.entry_end_time = Entry(self.panel_time,width=10)
        self.entry_end_time.insert(0,"-1")
        self.entry_end_time.pack(side="left")
        self.accum_plotmax = BooleanVar(self.panel_time,False)
        self.cb_accum_plotmax = Checkbutton(self.panel_time,text=_L["PLOT_MAX"],variable=self.accum_plotmax)
        self.cb_accum_plotmax.pack(side="left")
        self.panel_conf = Frame(self.lfra_head)
        self.panel_conf.pack(side="top",fill="x",anchor='w',pady=2,)
        self.lb_conf = Label(self.panel_conf, text=_L["FILE_EXT"])
        self.lb_conf.pack(side="left")
        self.cb_ext = Combobox(self.panel_conf,width=5, state="readonly")
        self.cb_ext['values'] = ["png","jpg","pdf","eps","svg","tiff"]
        self.cb_ext.current(0)
        self.cb_ext.pack(side="left")
        self.lb_dpi = Label(self.panel_conf, text=_L["IMAGE_DPI"])
        self.lb_dpi.pack(side="left")
        self.entry_dpi = Combobox(self.panel_conf,width=5)
        self.entry_dpi['values'] = ['128', '192', '256', '300', '400', '600', '1200']
        self.entry_dpi.current(3)
        self.entry_dpi.pack(side="left")
        self.plot_title = BooleanVar(self.panel_conf,True)
        self.cb_accum_plotmax = Checkbutton(self.panel_conf,text=_L["PLOT_TITLE"],variable=self.plot_title)
        self.cb_accum_plotmax.pack(side="left")

        self.plot_fcs = BooleanVar(self, False)
        self.cb_fcs = Checkbutton(self, text=_L["FCS_TITLE"], variable=self.plot_fcs)
        self.cb_fcs.grid(row=1,column=0,padx=3,pady=5,sticky='w')
        self.panel_fcs = Frame(self, border=1, relief='groove')
        self.panel_fcs.grid(row=2,column=0,sticky="nsew",padx=(5,3),pady=(0,5))
        self.fcs_opts = OptionBox(self.panel_fcs, {
            "wcnt": (_L["FCS_NVEH"], True),
            "load": (_L["FCS_PC"], True),
            "price": (_L["FCS_PRICE"], False),
        })
        self.fcs_opts.pack(side='top',fill='x',padx=3)
        self.fcs_pad = PlotPad(self.panel_fcs, True)
        self.fcs_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_scs = BooleanVar(self, False)
        self.cb_scs = Checkbutton(self, text=_L["SCS_TITLE"], variable=self.plot_scs)
        self.cb_scs.grid(row=3,column=0,padx=3,pady=5,sticky='w')
        self.panel_scs = Frame(self, border=1, relief='groove')
        self.panel_scs.grid(row=4,column=0,sticky="nsew",padx=(5,3),pady=(0,5))
        self.scs_opts = OptionBox(self.panel_scs, {
            "wcnt": (_L["SCS_NVEH"], True), 
            "cload": (_L["SCS_PC"], True), 
            "dload": (_L["SCS_PD"], True), 
            "netload": (_L["SCS_PPURE"], True), 
            "v2gcap": (_L["SCS_PV2G"], True), 
            "pricebuy": (_L["SCS_PBUY"], False), 
            "pricesell": (_L["SCS_PSELL"], False), 
        }, lcnt = 4)
        self.scs_opts.pack(side='top',fill='x',padx=3)
        self.scs_pad = PlotPad(self.panel_scs, True)
        self.scs_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_ev = BooleanVar(self, False)
        self.cb_ev = Checkbutton(self, text=_L["EV_TITLE"], variable=self.plot_ev)
        self.cb_ev.grid(row=1,column=1,padx=3,pady=5,sticky='w')
        self.panel_ev = Frame(self, border=1, relief='groove')
        self.panel_ev.grid(row=2,column=1,sticky="nsew",padx=(5,3),pady=(0,5))
        self.ev_opts = OptionBox(self.panel_ev, {
            "soc": (_L["SOC"], True),
            "status": (_L["EV_STA"], False),
            "cost": (_L["EV_COST"], True),
            "earn": (_L["EV_EARN"], True),
            "cpure": (_L["EV_NETCOST"], True),
        })
        self.ev_opts.pack(side='top',fill='x',padx=3)
        self.ev_pad = PlotPad(self.panel_ev, useEntry=True)
        self.ev_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_bus = BooleanVar(self, False)
        self.cb_bus = Checkbutton(self, text=_L["BUS_TITLE"], variable=self.plot_bus)
        self.cb_bus.grid(row=3,column=1,padx=3,pady=5,sticky='w')
        self.panel_bus = Frame(self, border=1, relief='groove')
        self.panel_bus.grid(row=4,column=1,sticky="nsew",padx=(5,3),pady=(0,5))
        self.bus_opts = OptionBox(self.panel_bus, {
            "activel": (_L["BUS_PD"], True),
            "reactivel": (_L["BUS_QD"], True),
            "volt": (_L["BUS_V"], True),
            "activeg": (_L["BUS_PG"], True),
            "reactiveg": (_L["BUS_QG"], True),
        },lcnt=3)
        self.bus_opts.pack(side='top',fill='x',padx=3)
        self.bus_pad = PlotPad(self.panel_bus, True, False, True)
        self.bus_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_gen = BooleanVar(self, False)
        self.cb_gen = Checkbutton(self, text=_L["GEN_TITLE"], variable=self.plot_gen)
        self.cb_gen.grid(row=5,column=0,padx=3,pady=5,sticky='w')
        self.panel_gen = Frame(self, border=1, relief='groove')
        self.panel_gen.grid(row=6,column=0,sticky="nsew",padx=(5,3),pady=(0,5))
        self.gen_opts = OptionBox(self.panel_gen, {
            "active": (_L["ACTIVE_POWER"], True),
            "reactive": (_L["REACTIVE_POWER"], True),
            "costp": (_L["GEN_COST"], True),
        })
        self.gen_opts.pack(side='top',fill='x',padx=3)
        self.gen_pad = PlotPad(self.panel_gen, True, False, True)
        self.gen_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_line = BooleanVar(self, False)
        self.cb_line = Checkbutton(self, text=_L["LINE_TITLE"], variable=self.plot_line)
        self.cb_line.grid(row=5,column=1,padx=3,pady=5,sticky='w')
        self.panel_line = Frame(self, border=1, relief='groove')
        self.panel_line.grid(row=6,column=1,sticky="nsew",padx=(5,3),pady=(0,5))
        self.line_opts = OptionBox(self.panel_line, {
            "active": (_L["ACTIVE_POWER"], True),
            "reactive": (_L["REACTIVE_POWER"], True),
            "current": (_L["LINE_CURRENT"], True),
        })
        self.line_opts.pack(side='top',fill='x',padx=3)
        self.line_pad = PlotPad(self.panel_line)
        self.line_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_pvw = BooleanVar(self, False)
        self.cb_pvw = Checkbutton(self, text=_L["PVW_TITLE"], variable=self.plot_pvw)
        self.cb_pvw.grid(row=7,column=0,padx=3,pady=5,sticky='w')
        self.panel_pvw = Frame(self, border=1, relief='groove')
        self.panel_pvw.grid(row=8,column=0,sticky="nsew",padx=(5,3),pady=(0,5))
        self.pvw_opts = OptionBox(self.panel_pvw, {
            "P": (_L["ACTIVE_POWER"], True),
            "cr": (_L["PVW_CR"], True),
        })
        self.pvw_opts.pack(side='top',fill='x',padx=3)
        self.pvw_pad = PlotPad(self.panel_pvw)
        self.pvw_pad.pack(side='top',fill='x',padx=3,pady=(0,3))

        self.plot_ess = BooleanVar(self, False)
        self.cb_ess = Checkbutton(self, text=_L["ESS_TITLE"], variable=self.plot_ess)
        self.cb_ess.grid(row=7,column=1,padx=3,pady=5,sticky='w')
        self.panel_ess = Frame(self, border=1, relief='groove')
        self.panel_ess.grid(row=8,column=1,sticky="nsew",padx=(5,3),pady=(0,5))
        self.ess_opts = OptionBox(self.panel_ess, {
            "P": (_L["ACTIVE_POWER"], True),
            "soc": (_L["SOC"], True),
        })
        self.ess_opts.pack(side='top',fill='x',padx=3)
        self.ess_pad = PlotPad(self.panel_ess)
        self.ess_pad.pack(side='top',fill='x',padx=3,pady=(0,3))
    
    def getConfig(self):
        return {
            "btime": int(self.entry_time.get()),
            "etime": int(self.entry_end_time.get()),
            "plotmax": self.accum_plotmax.get(),
            "fcs_accum": self.fcs_pad.accum.get() and self.plot_fcs.get(),
            "scs_accum": self.scs_pad.accum.get() and self.plot_scs.get(),
            "bus_total": self.bus_pad.accum.get() and self.plot_bus.get(),
            "gen_total": self.gen_pad.accum.get() and self.plot_gen.get(),
            "fcs": self.fcs_opts.getValues() if self.plot_fcs.get() else None,
            "scs": self.scs_opts.getValues() if self.plot_scs.get() else None,
            "ev": self.ev_opts.getValues() if self.plot_ev.get() else None,
            "gen": self.gen_opts.getValues() if self.plot_gen.get() else None,
            "bus": self.bus_opts.getValues() if self.plot_bus.get() else None,
            "line": self.line_opts.getValues() if self.plot_line.get() else None,
            "pvw": self.pvw_opts.getValues() if self.plot_pvw.get() else None,
            "ess": self.ess_opts.getValues() if self.plot_ess.get() else None,
        }

    def getTime(self):
        return int(self.entry_time.get()), int(self.entry_end_time.get())
    
    def pars(self, key:str):
        ret = self.getConfig()[key]
        assert isinstance(ret, dict), f"{key} is not a dict: {ret}"
        ret.update({
            "tl": int(self.entry_time.get()),
            "tr": int(self.entry_end_time.get())
        })
        return ret

    def enable(self, items:Optional[List[str]]=None):
        if items is None:
            items = AVAILABLE_ITEMS
        else:
            for i in items:
                assert i in AVAILABLE_ITEMS
        for i in items:
            getattr(self, f"cb_{i}")['state']=NORMAL
            getattr(self, f"{i}_opts").enable()
            getattr(self, f"{i}_pad").enable()
    
    def disable(self, items:List[str]=[]):
        if len(items)==0:
            items = AVAILABLE_ITEMS
        else:
            for i in items:
                assert i in AVAILABLE_ITEMS
        for i in items:
            getattr(self, f"cb_{i}")['state']=DISABLED
            getattr(self, f"{i}_opts").disable()
            getattr(self, f"{i}_pad").disable()


from pathlib import Path
from typing import Dict, List, Literal, Optional, Union, Iterable, Tuple
import bisect
from warnings import warn
import matplotlib
matplotlib.use("agg")
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.ticker import MultipleLocator, ScalarFormatter
from feasytools import SegFunc
from .reader import *
from ..locale import Lang

PLOT_ALL_CHARGE = Lang.PLOT_STR_ALL
PLOT_FAST_CHARGE = Lang.PLOT_STR_FAST
PLOT_SLOW_CHARGE = Lang.PLOT_STR_SLOW
DEFAULT_FONT = Lang.PLOT_FONT
FONT_SIZE_SMALL = int(Lang.PLOT_FONT_SIZE_SMALL)
FONT_SIZE_NORMAL = int(Lang.PLOT_FONT_SIZE_MEDIUM)
FONT_SIZE_LARGE = int(Lang.PLOT_FONT_SIZE_LARGE)

def split_string_except_quotes(string:str,delim:str,trim_quotes:bool=True) -> List[str]:
    result = []
    current_word = ""
    inside_quotes = False
    for char in string:
        if char in delim and not inside_quotes:
            if current_word:
                result.append(current_word)
                current_word = ""
        elif char == '"':
            inside_quotes = not inside_quotes
            if not trim_quotes:
                current_word += char
        else:
            current_word += char

    if current_word:
        result.append(current_word)

    return result

class AdvancedPlot:
    def __init__(self,tl:int=0,tr:int=-1,w:int=12,h:int=3,remove_edge:bool=True,double_side:bool=False,
            pic_ext:str="png",dpi:int=128,quick_plot_title:bool = True):
        self.__series:Dict[str,ReadOnlyStatistics] = {}
        self.fig = None
        self.dpi = dpi
        self.pic_ext = pic_ext
        self.plot_title = quick_plot_title
        self.new_fig(tl,tr,w,h,remove_edge,double_side)
        self.max_tr = -1

    def new_fig(self,tl:int=0,tr:int=-1,w:int=12,h:int=3,remove_edge:bool=True,double_side:bool=False):
        self.tl = tl
        self.tr = tr
        if self.fig is not None: plt.close(self.fig)
        self.fig, self.ax = plt.subplots(figsize=(w, h), dpi = self.dpi, constrained_layout=True)
        self.ax: Axes
        if remove_edge:
            self.ax.spines["right"].set_visible(False)
            self.ax.spines["top"].set_visible(False)
        if double_side:
            self.ax2 = self.ax.twinx()
        else:
            self.ax2 = None
    
    def check_right(self):
        if self.ax2 is None:
            self.ax2 = self.ax.twinx()

    def calc_expr(self, expr:str):
        i = 0
        obj = ""
        vars_dics = {}
        ret_expr = ""
        rb = -1
        lb = expr.find('{')
        lp = 0
        while lb != -1:
            rb = expr.find('}',lb)
            if rb == -1:
                raise ValueError(f"Unmatched bracket in expression '{expr}'")
            obj = expr[lb+1:rb]
            if obj == "": raise ValueError("Empty variable name")
            val = self.get_series(obj)
            vars_dics[f"v{i}"] = val
            ret_expr += expr[lp:lb] + f"v{i}"
            lp = rb + 1
            lb = expr.find('{',rb)
            i += 1
        if expr.find('}',lp) != -1:
            raise ValueError(f"Exccessive right bracket in expression '{expr}'")
        if lp < len(expr):
            ret_expr += expr[lp:]
        return eval(ret_expr.replace("^","**"),vars_dics)
    
    def get_accum_series(self, series:str) -> Tuple[List[int],List[List[Any]],List[str]]:
        s = series.split("|")
        if len(s) == 2:
            path,domain = s
            tl = self.tl
            tr = self.tr
        elif len(s) == 4:
            path,domain,tl,tr = s
            tl = int(tl)
            tr = int(tr)
        else:
            raise ValueError(f"Unsupported series '{series}': Should be 'path:domain:value' or 'path:domain:value:tl:tr'")
        if path not in self.__series:
            try:
                t = ReadOnlyStatistics(path)
                self.__series[path] = t
                self.max_tr = max(self.max_tr, t.LastTime)
            except:
                raise RuntimeError(f"Fail to load results directory '{path}'")
        se = self.__series[path]
        if domain == "fcs_load":
            d = se.FCS_load_all(tl,tr)
        elif domain == "scs_load":
            d = se.SCS_net_load_all(tl,tr)
        else:
            raise ValueError(f"Unsupported domain '{domain}'")
        x,y = SegFunc.cross_interpolate(d)
        return x,y,se.FCS_head if domain.startswith("fcs") else se.SCS_head

    def load_series(self, path:Union[str, ReadOnlyStatistics]):
        if isinstance(path, str):
            try:
                t = ReadOnlyStatistics(path)
                self.__series[path] = t
                self.max_tr = max(self.max_tr, t.LastTime)
            except:
                raise RuntimeError(f"Fail to load results directory '{path}'")
        else:
            self.__series[path.root] = path
            self.max_tr = max(self.max_tr, path.LastTime)
    
    def get_series(self, series:str) -> SegFunc:
        s = series.split("|")
        if len(s) == 2:
            path,domain = s
            val = ""
            tl = self.tl
            tr = self.tr
        elif len(s) == 3:
            path,domain,val = s
            tl = self.tl
            tr = self.tr
        elif len(s) == 4:
            path,domain,tl,tr = s
            tl = int(tl)
            tr = int(tr)
            val = ""
        elif len(s) == 5:
            path,domain,val,tl,tr = s
            tl = int(tl)
            tr = int(tr)
        else:
            raise ValueError(f"Unsupported series '{series}': Should be 'path|domain|value' or 'path|domain|value|tl|tr'")
        if path not in self.__series:
            try:
                t = ReadOnlyStatistics(path)
                self.__series[path] = t
                self.max_tr = max(self.max_tr, t.LastTime)
            except:
                raise RuntimeError(f"Fail to load results directory '{path}'")
        se = self.__series[path]
        if domain == "gen_total_active":
            d = se.G_total("totP")
        elif domain == "gen_total_reactive":
            d = se.G_total("totQ")
        elif domain == "gen_total_costp":
            d = se.G_total("totC")
        elif domain == "bus_total_active_load":
            d = se.bus_total("totPd")
        elif domain == "bus_total_reactive_load":
            d = se.bus_total("totQd")
        elif domain == "bus_total_active_gen":
            d = se.bus_total("totPg")
        elif domain == "bus_total_reactive_gen":
            d = se.bus_total("totQg")
        else:
            if val == "": raise ValueError("Value cannot be empty! Orginal series: "+series)
            if domain == "fcs_load":
                d = se.FCS_load_of(val)
            elif domain == "fcs_count":
                d = se.FCS_count_of(val)
            elif domain == "fcs_price_buy":
                d = se.FCS_pricebuy_of(val)
            elif domain == "scs_cload":
                d = se.SCS_charge_load_of(val)
            elif domain == "scs_dload":
                d = se.SCS_v2g_load_of(val)
            elif domain == "scs_load":
                d = se.SCS_net_load_of(val)
            elif domain == "scs_count":
                d = se.SCS_count_of(val)     
            elif domain == "scs_price_buy":
                d = se.SCS_pricebuy_of(val)
            elif domain == "scs_price_sell":
                d = se.SCS_pricesell_of(val)
            elif domain == "scs_vcap":
                d = se.SCS_v2g_cap_of(val)
            elif domain == "ev_soc":
                d = se.EV_attrib_of(val,"soc")
            elif domain == "ev_cost":
                d = se.EV_attrib_of(val,"cost")
            elif domain == "ev_earn":
                d = se.EV_attrib_of(val,"earn")
            elif domain == "ev_cpure":
                d = se.EV_attrib_of(val,"cost") - se.EV_attrib_of(val,"earn")
            elif domain == "ev_status":
                d = se.EV_attrib_of(val,"status")
            elif domain == "gen_active":
                d = se.G_attrib_of(val,"P")
            elif domain == "gen_reactive":
                d = se.G_attrib_of(val,"Q")
            elif domain == "gen_costp":
                d = se.G_attrib_of(val,"costp")
            elif domain == "bus_voltage":
                d = se.bus_attrib_of(val,"V")
            elif domain == "bus_active_load":
                d = se.bus_attrib_of(val,"Pd")
            elif domain == "bus_reactive_load":
                d = se.bus_attrib_of(val,"Qd")
            elif domain == "bus_active_gen":
                d = se.bus_attrib_of(val,"Pg")
            elif domain == "bus_reactive_gen":
                d = se.bus_attrib_of(val,"Qg")
            elif domain == "line_active":
                d = se.line_attrib_of(val,"P")
            elif domain == "line_reactive":
                d = se.line_attrib_of(val,"Q")
            elif domain == "line_current":
                d = se.line_attrib_of(val,"I")
            elif domain == "pvw_p":
                d = se.pvw_attrib_of(val,"P")
            elif domain == "pvw_cr":
                d = se.pvw_attrib_of(val,"curt")
            elif domain == "ess_p":
                d = se.ess_attrib_of(val,"P")
            elif domain == "ess_soc":
                d = se.ess_attrib_of(val,"soc")
            else:
                raise ValueError(f"Unsupported domain '{domain}'")
        return d.slice(tl, self.max_tr).interpolate(tl, self.max_tr)
    
    def add_data(self, 
            expr:str, 
            label:Optional[str]=None, 
            color:Optional[str]=None, 
            linestyle:Optional[str]="-",
            linewidth:Optional[float]=1,
            side:Literal["left","right"]="left"
        ):
        kwargs = {}
        if label is not None:
            kwargs["label"] = label
        if color is not None:
            kwargs["color"] = color
        if linestyle is not None:
            kwargs["linestyle"] = linestyle
        if linewidth is not None:
            kwargs["linewidth"] = linewidth
        data = self.calc_expr(expr)
        assert isinstance(data, SegFunc)
        if side == "left":
            self.ax.plot(data.time, data.data, **kwargs)
        else:
            self.check_right()
            assert isinstance(self.ax2, Axes)
            self.ax2.plot(data.time, data.data, **kwargs)
        self.__setticks()
    
    def __setticks(self):
        tl = self.tl
        tr = self.max_tr if self.tr == -1 else self.tr
        if tl % 60 != 0: tl -= tl%60
        if tr % 60 != 0: tr += (60-tr%60)
        k = (tr - tl) // 12
        ava_k = [900, 1800, 3600, 7200, 3*3600, 4*3600,6*3600, 8*3600, 12*3600, 24*3600]
        p = bisect.bisect_right(ava_k, k) - 1
        if p<0: p = 0
        k = ava_k[p]
        def __trans(i:int):
            return f"D{i//86400}-{(i%86400)//3600:02}:{(i%3600)//60:02}"
        self.ax.set_xticks(
            list(range(tl, tr+1, k)),
            map(__trans, range(tl, tr+1, k))
        )
        plt.setp(self.ax.get_xticklabels(), rotation=20, ha="right", rotation_mode="anchor")
    
    def add_accum(self,name:str,plot_max:bool):
        x,y,lbs = self.get_accum_series(name)
        self.ax.stackplot(x,y,labels=lbs)
        if plot_max:
            max_y = max([sum(y[i][j] for i in range(len(y))) for j in range(len(x))])
            print(max_y)
            self.ax.plot([x[0],x[-1]],[max_y,max_y],color="black",linestyle="--",linewidth=1.5)
            self.ax.text(x[0] + (x[-1] - x[0]) * 0.01, max_y * 0.9, f"Max = {max_y:.1f}kW", fontsize = FONT_SIZE_SMALL, color="black")
        self.__setticks()
    
    def y_label(self,label:str,side:Literal["left","right"]="left"):
        if side == "left":
            self.yleft_label(label)
        else:
            self.yright_label(label)
    
    def yleft_label(self,label:str):
        self.ax.set_ylabel(label, fontweight="bold", font=DEFAULT_FONT, fontsize=FONT_SIZE_NORMAL)
    
    def yleft_range(self,low:float,high:float):
        self.ax.set_ylim(low, high)

    def yright_label(self,label:str):
        self.check_right()
        assert isinstance(self.ax2, Axes)
        self.ax2.set_ylabel(label, fontweight="bold", font=DEFAULT_FONT, fontsize=FONT_SIZE_NORMAL)
    
    def yright_range(self,low:float,high:float):
        self.check_right()
        assert isinstance(self.ax2, Axes)
        self.ax2.set_ylim(low,high)
    
    def x_label(self,label:str):
        self.ax.set_xlabel(label, fontweight="bold", font=DEFAULT_FONT, fontsize=FONT_SIZE_NORMAL)
    
    def title(self,title:str):
        self.ax.set_title(title, font=DEFAULT_FONT, fontsize=FONT_SIZE_LARGE)
    
    def yticks(self,ticks,labels:Optional[Iterable[str]] = None):
        self.ax.set_yticks(ticks,labels)

    def legend(self,loc:str="upper right",ncol:int=1):
        assert self.fig is not None
        if len(self.ax.get_legend_handles_labels()[1]) > 100:
            warn("Too many legend entries (>100). The legend will not be shown.")
        else:
            self.fig.legend(loc=loc, bbox_to_anchor=(1, 1), bbox_transform=self.ax.transAxes,ncol=ncol)

    def hline(self,y:float,color:str="black",linestyle:str="--",linewidth:float=1.5):
        self.ax.axhline(y=y,color=color,linestyle=linestyle,linewidth=linewidth)
    
    def vline(self,x:float,color:str="black",linestyle:str="--",linewidth:float=1.5):
        self.ax.axvline(x=x,color=color,linestyle=linestyle,linewidth=linewidth)
    
    def save(self,save_to:str,mloc:bool=True,zeroxmargin:bool=True):
        if mloc:
            x_major_locator = MultipleLocator(3600)
            self.ax.xaxis.set_major_locator(x_major_locator)
        if zeroxmargin:
            self.ax.margins(x=0)
        y_formatter = ScalarFormatter(useOffset=False)
        self.ax.yaxis.set_major_formatter(y_formatter)
        self.ax.set_xlim(self.tl, self.max_tr if self.tr == -1 else self.tr)
        assert self.fig is not None
        self.fig.savefig(save_to, dpi=self.dpi, bbox_inches="tight")
        plt.close()
    
    def configure(self,commands:Union[str, List[str]]):
        if isinstance(commands,str):
            commands = split_string_except_quotes(commands,";",False)
        else:
            new_cmds = []
            for cmd in commands:
                new_cmds.extend(split_string_except_quotes(cmd,";",False))
            commands = new_cmds
        for cmd in commands:
            cmds = split_string_except_quotes(cmd.strip()," ,")
            if len(cmds) == 0:
                continue
            if cmds[0] == "plot":
                if len(cmds) <= 1:
                    raise ValueError("Series not provided")
                expr = cmds[1] 
                kwargs = {
                    "label": cmds[2] if len(cmds) > 2 else None,
                    "color": cmds[3] if len(cmds) > 3 else None,
                    "linestyle": cmds[4] if len(cmds) > 4 else "-",
                    "linewidth": float(cmds[5]) if len(cmds) > 5 else 1,
                    "side": cmds[6] if len(cmds) > 6 else "left"
                }
                self.add_data(expr,**kwargs)
            elif cmds[0] == "plotaccum":
                if len(cmds) <= 1:
                    raise ValueError("Series not provided")
                plot_max = cmds[2] == "True" if len(cmds) > 2 else False
                self.add_accum(cmds[1], plot_max)
            elif cmds[0] == "new":
                if len(cmds) <= 2:
                    raise ValueError("width and height not provided")
                self.fig, self.ax = plt.subplots(
                    figsize=(int(cmds[1]), int(cmds[2])), 
                    dpi=128, 
                    constrained_layout=True
                )
            elif cmds[0] == "title":
                self.title(cmds[1])
                if len(cmds) <= 1:
                    raise ValueError("Title not provided")
            elif cmds[0] == "xlabel":
                self.x_label(cmds[1])
                if len(cmds) <= 1:
                    raise ValueError("X-axis label not provided")
            elif cmds[0] == "yleftlabel" or cmds[0] == "ylabel":
                self.yleft_label(cmds[1])
                if len(cmds) <= 1:
                    raise ValueError("Y-axis label not provided")
            elif cmds[0] == "yrightlabel":
                self.yright_label(cmds[1])
                if len(cmds) <= 1:
                    raise ValueError("Y-axis label not provided")
            elif cmds[0] == "yticks":
                if len(cmds) <= 1:
                    raise ValueError("Y-axis ticks not provided")
                self.yticks(
                    [float(x) for x in cmds[1].split(",")],
                    cmds[2].split(",") if len(cmds) > 2 else None
                )
            elif cmds[0] == "legend":
                self.legend(
                    cmds[1] if len(cmds) > 1 else "upper right", 
                    int(cmds[2]) if len(cmds) > 2 else 1
                )
            elif cmds[0] == "save":
                self.save(cmds[1])
            elif cmds[0] == "default_trunc":
                assert len(cmds) == 3
                self.tl = int(cmds[1])
                self.tr = int(cmds[2])
                self.max_tr = self.tr
            elif cmds[0] == "hline":
                self.hline(
                    float(cmds[1]),
                    cmds[2] if len(cmds) > 2 else "black",
                    cmds[3] if len(cmds) > 3 else "--",
                    float(cmds[4]) if len(cmds) > 4 else 1.5
                )
            elif cmds[0] == "vline":
                self.vline(
                    float(cmds[1]),
                    cmds[2] if len(cmds) > 2 else "black",
                    cmds[3] if len(cmds) > 3 else "--",
                    float(cmds[4]) if len(cmds) > 4 else 1.5
                )
            elif cmds[0] == "exit":
                return False
            elif cmds[0] == "help":
                print(Lang.ADV_PLOT_HELP)
            elif cmds[0] == "":
                continue
            else:
                raise ValueError(f"Unsupported command '{cmds[0]}'")
        return True
    
    def quick_fcs_accum(self,tl:int,tr:int,plot_max:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        self.add_accum(f"{res_path}|fcs_load", plot_max)
        self.yleft_label(Lang.PLOT_YLABEL_POWERKW)
        if self.plot_title: self.title(Lang.PLOT_FCS_ACC_TITLE)
        self.legend(ncol=2)
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if save_to == "": save_to = str(p / f"fcs_total.{self.pic_ext}")
        self.save(save_to)
    
    def quick_scs_accum(self,tl:int,tr:int,plot_max:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        self.add_accum(f"{res_path}|scs_load", plot_max)
        self.yleft_label(Lang.PLOT_YLABEL_POWERKW)
        if self.plot_title: self.title(Lang.PLOT_SCS_ACC_TITLE)
        self.legend(ncol=2)
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if save_to == "": save_to = str(p / f"scs_total.{self.pic_ext}")
        self.save(save_to)
    
    def quick_fcs(self,tl:int,tr:int,cs_name:str,load:bool,price:bool,wcnt:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if cs_name == "":
            raise ValueError("CS name cannot be empty")
        if load:
            self.add_data("{"+f"{res_path}|fcs_load|{cs_name}"+"}","Load",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERKW)
        if price:
            side = "right" if load else "left"
            self.add_data("{"+f"{res_path}|fcs_price_buy|{cs_name}"+"}","Price",color="red",side=side)
            self.y_label(Lang.PLOT_YLABEL_PRICE, side)
        if wcnt:
            if load and price:
                raise RuntimeError("Cannot plot count with load and price, since there are only 2 sides")
            side = "right" if load or price else "left"
            self.add_data("{"+f"{res_path}|fcs_count|{cs_name}"+"}","Veh. Count",color="green",side=side)
            self.y_label(Lang.PLOT_YLABEL_COUNT,side)
        if self.plot_title: self.title(Lang.PLOT_FCS_TITLE.format(cs_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if cs_name == "<sum>": cs_name = "sum"
        if save_to == "": save_to = str(p / f"fcs_{cs_name}.{self.pic_ext}")
        self.save(save_to)
    
    def quick_scs(self,tl:int,tr:int,cs_name:str,cload:bool,dload:bool,netload:bool,v2gcap:bool,wcnt:bool,pricebuy:bool,pricesell:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        has_load = cload or dload or netload or v2gcap
        if has_load:
            self.yleft_label(Lang.PLOT_YLABEL_POWERKW)
            if cload:
                self.add_data("{"+f"{res_path}|scs_cload|{cs_name}"+"}","Charging Load",color="blue")
            if dload:
                self.add_data("{"+f"{res_path}|scs_dload|{cs_name}"+"}", "Discharging Load",color="red")
            if netload:
                self.add_data("{"+f"{res_path}|scs_load|{cs_name}"+"}", "Net Load",color="green")
            if v2gcap:
                self.add_data("{"+f"{res_path}|scs_vcap|{cs_name}"+"}", "V2G Capacity",color="purple")
        
        has_price = pricebuy or pricesell
        if has_price:
            side = "right" if has_load else "left"
            if pricebuy:
                self.add_data("{"+f"{res_path}|scs_price_buy|{cs_name}"+"}","Price Buy",color="orange",side=side)
            if pricesell:
                self.add_data("{"+f"{res_path}|scs_price_sell|{cs_name}"+"}","Price Sell",color="cyan",side=side)
            self.y_label(Lang.PLOT_YLABEL_PRICE, side)
            
        if wcnt:
            if has_load and has_price:
                raise RuntimeError("Cannot plot count with load and price, since there are only 2 sides")
            side = "right" if has_load or has_price else "left"
            self.add_data("{"+f"{res_path}|scs_count|{cs_name}"+"}","Veh. Count",color="gray",side=side)
            self.y_label(Lang.PLOT_YLABEL_COUNT, side)
        if self.plot_title: self.title(Lang.PLOT_SCS_TITLE.format(cs_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if cs_name == "<sum>": cs_name = "sum"
        if save_to == "": save_to = str(p / f"scs_{cs_name}.{self.pic_ext}")
        self.save(save_to)
    
    def quick_bus(self,tl:int,tr:int,bus_name:str,activel:bool,reactivel:bool,volt:bool,
                  activeg:bool,reactiveg:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if volt:
            self.add_data("{"+f"{res_path}|bus_voltage|{bus_name}"+"}","Voltage",color="green")
            self.yleft_label(Lang.PLOT_YLABEL_VOLTKV)
        if activel:
            side = "right" if volt else "left"
            self.add_data("{"+f"{res_path}|bus_active_load|{bus_name}"+"}","Active Load",color="blue",side=side)
            self.y_label(Lang.PLOT_YLABEL_POWERMW, side)
        if reactivel:
            side = "right" if volt else "left"
            self.add_data("{"+f"{res_path}|bus_reactive_load|{bus_name}"+"}","Reactive Load",color="red",side=side)
            self.y_label(Lang.PLOT_YLABEL_POWERMW, side)
        if activeg:
            side = "right" if volt else "left"
            self.add_data("{"+f"{res_path}|bus_active_gen|{bus_name}"+"}","Active Gen",color="purple",side=side)
            self.y_label(Lang.PLOT_YLABEL_POWERMW, side)
        if reactiveg:
            side = "right" if volt else "left"
            self.add_data("{"+f"{res_path}|bus_reactive_gen|{bus_name}"+"}","Reactive Gen",color="cyan",side=side)
            self.y_label(Lang.PLOT_YLABEL_POWERMW, side)
        if self.plot_title: self.title(Lang.PLOT_BUS.format(bus_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if bus_name == "<sum>": bus_name = "sum"
        if save_to == "": save_to = str(p / f"bus_{bus_name}.{self.pic_ext}")
        self.save(save_to)
        
    def quick_gen(self,tl:int,tr:int,gen_name:str,active:bool,reactive:bool,costp:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if active:
            self.add_data("{"+f"{res_path}|gen_active|{gen_name}"+"}","Active Power",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERMW)
        if reactive:
            self.add_data("{"+f"{res_path}|gen_reactive|{gen_name}"+"}","Reactive Power",color="red")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if costp:
            side = "right" if active or reactive else "left"
            self.add_data("{"+f"{res_path}|gen_costp|{gen_name}"+"}","Cost",color="green",side=side)
            self.y_label(Lang.PLOT_YLABEL_COST, side)
        if self.plot_title: self.title(Lang.PLOT_GEN.format(gen_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if gen_name == "<sum>": gen_name = "sum"
        if save_to == "": save_to = str(p / f"gen_{gen_name}.{self.pic_ext}")
        self.save(save_to)
    
    def quick_ev(self,tl:int,tr:int,ev_name:str,soc:bool,status:bool,cost:bool,earn:bool,cpure:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if soc:
            self.add_data("{"+f"{res_path}|ev_soc|{ev_name}"+"}","SOC",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_SOC)
        if cost:
            side = "right" if soc else "left"
            self.add_data("{"+f"{res_path}|ev_cost|{ev_name}"+"}","Cost",color="green",side=side)
            self.y_label(Lang.PLOT_YLABEL_COST, side)
        if earn:
            side = "right" if soc else "left"
            self.add_data("{"+f"{res_path}|ev_earn|{ev_name}"+"}","Earn",color="purple",side=side)
            self.y_label(Lang.PLOT_YLABEL_COST, side)
        if cpure:
            side = "right" if soc else "left"
            self.add_data("{"+f"{res_path}|ev_cpure|{ev_name}"+"}","Pure cost",color="cyan",side=side)
            self.y_label(Lang.PLOT_YLABEL_COST, side)
        if status:
            if soc and (cost or earn or cpure):
                raise RuntimeError("Cannot plot cost with soc and status, since there are only 2 sides")
            side = "right" if soc or cost or earn or cpure else "left"
            self.add_data("{"+f"{res_path}|ev_status|{ev_name}"+"}","Status",color="red")
            self.y_label(Lang.PLOT_YLABEL_STATUS)
        if cost and earn:
            self.legend(ncol=2)
        else:
            self.legend()
        if self.plot_title: self.title(Lang.PLOT_EV.format(ev_name))
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if save_to == "": save_to = str(p / f"ev_{ev_name}.{self.pic_ext}")
        self.save(save_to)
        
    def quick_gen_tot(self,tl:int,tr:int,active:bool,reactive:bool,costp:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if active:
            self.add_data("{"+f"{res_path}|gen_total_active|<sum>"+"}","Active Power",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERMW)
        if reactive:
            self.add_data("{"+f"{res_path}|gen_total_reactive|<sum>"+"}","Reactive Power",color="red")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if costp:
            side = "right" if active or reactive else "left"
            self.add_data("{"+f"{res_path}|gen_total_costp|<sum>"+"}","Cost",color="green",side=side)
            self.y_label(Lang.PLOT_YLABEL_COST,side)
        if self.plot_title: self.title(Lang.PLOT_GEN_TOTAL)
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if save_to == "": save_to = str(p / f"gen_total.{self.pic_ext}")
        self.save(save_to)
    
    def quick_bus_tot(self,tl:int,tr:int,active:bool,reactive:bool,active_gen:bool,reactive_gen:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if active:
            self.add_data("{"+f"{res_path}|bus_total_active_load"+"}","Active Load",color="blue")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if reactive:
            self.add_data("{"+f"{res_path}|bus_total_reactive_load"+"}","Reactive Load",color="red")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if active_gen:
            self.add_data("{"+f"{res_path}|bus_total_active_gen"+"}","Active Gen",color="green")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if reactive_gen:
            self.add_data("{"+f"{res_path}|bus_total_reactive_gen"+"}","Reactive Gen",color="purple")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if self.plot_title: self.title(Lang.PLOT_BUS_TOTAL)
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if save_to == "": save_to = str(p / f"bus_total.{self.pic_ext}")
        self.save(save_to)
    
    def quick_line(self,tl:int,tr:int,line_name:str,active:bool,reactive:bool,current:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if active:
            self.add_data("{"+f"{res_path}|line_active|{line_name}"+"}","Active Power",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERMW)
        if reactive:
            self.add_data("{"+f"{res_path}|line_reactive|{line_name}"+"}","Reactive Power",color="red")
            self.y_label(Lang.PLOT_YLABEL_POWERMW)
        if current:
            side = "right" if active or reactive else "left"
            self.add_data("{"+f"{res_path}|line_current|{line_name}"+"}","Current",color="green",side=side)
            self.y_label(Lang.PLOT_YLABEL_CURRENT, side)
        if self.plot_title: self.title(Lang.PLOT_LINE.format(line_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if line_name == "<sum>": line_name = "sum"
        if save_to == "": save_to = str(p / f"line_{line_name}.{self.pic_ext}")
        self.save(save_to)
    
    def quick_pvw(self,tl:int,tr:int,pvw_name:str,P:bool,cr:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if P:
            self.add_data("{"+f"{res_path}|pvw_p|{pvw_name}"+"}","Active Power",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERMW)
        if cr:
            side = "right" if P else "left"
            self.add_data("{"+f"{res_path}|pvw_cr|{pvw_name}"+"}","Curtailment Rate",color="red")
            self.y_label(Lang.PLOT_YLABEL_CURTAIL, side=side)
        if self.plot_title: self.title(Lang.PLOT_PVW.format(pvw_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if pvw_name == "<sum>": pvw_name = "sum"
        if save_to == "": save_to = str(p / f"pvw_{pvw_name}.{self.pic_ext}")
        self.save(save_to)
    
    def quick_ess(self,tl:int,tr:int,ess_name:str,P:bool,soc:bool,save_to:str="",res_path:str="results"):
        self.new_fig(tl,tr)
        if P:
            self.add_data("{"+f"{res_path}|ess_p|{ess_name}"+"}","Active Power",color="blue")
            self.yleft_label(Lang.PLOT_YLABEL_POWERMW)
        if soc:
            side = "right" if P else "left"
            self.add_data("{"+f"{res_path}|ess_soc|{ess_name}"+"}","SoC",color="red")
            self.y_label(Lang.PLOT_YLABEL_SOC, side=side)
        if self.plot_title: self.title(Lang.PLOT_ESS.format(ess_name))
        self.legend()
        self.x_label(Lang.PLOT_XLABEL_TIME)
        p = Path(res_path) / "figures"
        p.mkdir(parents=True,exist_ok=True)
        if ess_name == "<sum>": ess_name = "sum"
        if save_to == "": save_to = str(p / f"ess_{ess_name}.{self.pic_ext}")
        self.save(save_to)


from v2simux_gui.common import *

from v2simux import CS, SCS, FCS, ListSelection, PricingMethod, LoadCSList
from feasytools import SegFunc, ConstFunc, OverrideFunc, RangeList
from .utils import *
from .controls import ScrollableTreeView, ALWAYS_ONLINE

_L = LangLib.Load(__file__)

class CSEditorGUI(Frame):
    def __init__(self, master, generatorFunc, canV2g:bool, file:str="", **kwargs):
        super().__init__(master, **kwargs)

        self._Q = EventQueue(self)
        self._Q.register("loaded", lambda: None)
        
        self.gf = generatorFunc
        if file:
            self.file = file
        else:
            self.file = ""
        
        self.tree = ScrollableTreeView(self, allowSave=True) 
        self.tree['show'] = 'headings'
        if canV2g:
            self.csType = SCS
            self.tree["columns"] = ("Node", "Slots", "Bus", "x", "y", "Online", "MaxPc", "MaxPd", "PriceBuy", "PriceSell", "PcAlloc", "PdAlloc")
        else:
            self.csType = FCS
            self.tree["columns"] = ("Node", "Slots", "Bus", "x", "y", "Online", "MaxPc", "PriceBuy", "PcAlloc")
        self.tree.column("Node", width=120, stretch=NO)
        self.tree.column("Slots", width=90, stretch=NO)
        self.tree.column("Bus", width=80, stretch=NO)
        self.tree.column("x", width=60, stretch=NO)
        self.tree.column("y", width=60, stretch=NO)
        self.tree.column("Online", width=100, stretch=NO)
        self.tree.column("MaxPc", width=130, stretch=NO)
        self.tree.column("PriceBuy", width=120, stretch=YES)
        if canV2g:
            self.tree.column("MaxPd", width=130, stretch=NO)
            self.tree.column("PriceSell", width=120, stretch=YES)
            self.tree.column("PcAlloc", width=80, stretch=NO)
            self.tree.column("PdAlloc", width=80, stretch=NO)
        
        self.tree.heading("Node", text=_L["CSE_NODE"])
        self.tree.heading("Slots", text=_L["CSE_SLOTS"])
        self.tree.heading("Bus", text=_L["CSE_BUS"])
        self.tree.heading("x", text=_L["CSE_X"])
        self.tree.heading("y", text=_L["CSE_Y"])
        self.tree.heading("Online", text=_L["CSE_OFFLINE"])
        self.tree.heading("MaxPc", text=_L["CSE_MAXPC"])
        self.tree.heading("PriceBuy", text=_L["CSE_PRICEBUY"])
        self.tree.heading("PcAlloc", text=_L["CSE_PCALLOC"])

        self.tree.setColEditMode("Node", EditMode.entry())
        self.tree.setColEditMode("Slots", EditMode.spin(0, 100))
        self.tree.setColEditMode("Bus", EditMode.entry())
        self.tree.setColEditMode("x", EditMode.entry())
        self.tree.setColEditMode("y", EditMode.entry())
        self.tree.setColEditMode("Online", EditMode.rangelist(hint=True))
        self.tree.setColEditMode("MaxPc", EditMode.spin(0, 1000))
        self.tree.setColEditMode("PriceBuy", EditMode.segfunc())
        self.tree.setColEditMode("PcAlloc", EditMode.combo(values=["Average", "Prioritized"]))

        if canV2g:
            self.tree.heading("PriceSell", text=_L["CSE_PRICESELL"])
            self.tree.heading("MaxPd", text=_L["CSE_MAXPD"])
            self.tree.heading("PdAlloc", text=_L["CSE_PDALLOC"])
            self.tree.setColEditMode("PriceSell", EditMode.segfunc())
            self.tree.setColEditMode("MaxPd", EditMode.spin(0, 1000))
            self.tree.setColEditMode("PdAlloc", EditMode.combo(values=["Average"]))
        self.tree.pack(fill="both", expand=True)

        self.panel2 = Frame(self)
        self.btn_find = Button(self.panel2, text=_L["BTN_FIND"], command=self._on_btn_find_click)
        self.btn_find.pack(fill="x", side='right', anchor='e', expand=False)
        self.entry_find = Entry(self.panel2)
        self.entry_find.pack(fill="x", side='right', anchor='e',expand=False)
        self.lb_cnt = Label(self.panel2, text=_L["LB_COUNT"].format(0))
        self.lb_cnt.pack(fill="x", side='left', anchor='w', expand=False)
        self.panel2.pack(fill="x")

        self.gens = LabelFrame(self, text=_L["CS_GEN"])
        self.gens.pack(fill="x", expand=False)

        self.useMode = IntVar(self, 0)
        self.group_use = LabelFrame(self.gens, text=_L["CS_MODE"])
        self.rb_useAll = Radiobutton(self.group_use, text=_L["CS_USEALL"], value=0, variable=self.useMode, command=self._useModeChanged)
        self.rb_useAll.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.rb_useSel = Radiobutton(self.group_use, text=_L["CS_SELECTED"], value=1, variable=self.useMode, command=self._useModeChanged)
        self.rb_useSel.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.entry_sel = Entry(self.group_use, state=DISABLED)
        self.entry_sel.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.rb_useRandN = Radiobutton(self.group_use, text=_L["CS_RANDOM"], value=2, variable=self.useMode, command=self._useModeChanged)
        self.rb_useRandN.grid(row=0,column=3,padx=3,pady=3,sticky="w")
        self.entry_randN = Entry(self.group_use, state=DISABLED)
        self.entry_randN.grid(row=0,column=4,padx=3,pady=3,sticky="w")
        self.group_use.grid(row=2,column=0,padx=3,pady=3,sticky="nesw")

        self.use_cscsv = IntVar(self, 0)
        self.group_src = LabelFrame(self.gens, text=_L["CS_SRC"])
        self.rb_rnet = Radiobutton(self.group_src, text=_L["CS_USENODES"], value=0, variable=self.use_cscsv)
        self.rb_rnet.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.rb_cscsv = Radiobutton(self.group_src, text=_L["CS_USECSV"], value=1, variable=self.use_cscsv, state=DISABLED)
        self.rb_cscsv.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.rb_poly = Radiobutton(self.group_src, text=_L["CS_USEPOLY"], value=2, variable=self.use_cscsv,state=DISABLED)
        self.rb_poly.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.group_src.grid(row=1,column=0,padx=3,pady=3,sticky="nesw")

        self.fr = Frame(self.gens)
        self.lb_slots = Label(self.fr, text=_L["CS_SLOTS"])
        self.lb_slots.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.entry_slots = Entry(self.fr)
        self.entry_slots.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.entry_slots.insert(0, "10")
        self.lb_seed = Label(self.fr, text=_L["CS_SEED"])
        self.lb_seed.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.entry_seed = Entry(self.fr)
        self.entry_seed.grid(row=0,column=3,padx=3,pady=3,sticky="w")
        self.entry_seed.insert(0, "0")
        self.fr.grid(row=0,column=0,padx=3,pady=3,sticky="nesw")
        
        self.pbuy = IntVar(self, 1)
        self.group_pbuy = LabelFrame(self.gens, text=_L["CS_PRICEBUY"])
        self.rb_pbuy0 = Radiobutton(self.group_pbuy, text=_L["CS_PB5SEGS"], value=0, variable=self.pbuy, command=self._pBuyChanged)
        self.rb_pbuy0.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.rb_pbuy1 = Radiobutton(self.group_pbuy, text=_L["CS_PBFIXED"], value=1, variable=self.pbuy, command=self._pBuyChanged)
        self.rb_pbuy1.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.entry_pbuy = Entry(self.group_pbuy)
        self.entry_pbuy.insert(0, "1.0")
        self.entry_pbuy.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.group_pbuy.grid(row=3,column=0,padx=3,pady=3,sticky="nesw")

        self.psell = IntVar(self, 1)
        self.group_psell = LabelFrame(self.gens, text=_L["CS_PRICESELL"])
        self.rb_psell0 = Radiobutton(self.group_psell, text=_L["CS_PB5SEGS"], value=0, variable=self.psell, command=self._pSellChanged)
        self.rb_psell0.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.rb_psell1 = Radiobutton(self.group_psell, text=_L["CS_PBFIXED"], value=1, variable=self.psell, command=self._pSellChanged)
        self.rb_psell1.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.entry_psell = Entry(self.group_psell)
        self.entry_psell.insert(0, "1.5")
        self.entry_psell.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.group_psell.grid(row=4,column=0,padx=3,pady=3,sticky="nesw")

        self.busMode = IntVar(self, 0)
        self.group_bus = LabelFrame(self.gens, text=_L["CS_BUSMODE"])
        self.rb_busGrid = Radiobutton(self.group_bus, text=_L["CS_BUSBYPOS"], value=0, variable=self.busMode, command=self._busModeChanged)
        self.rb_busGrid.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.rb_busAll = Radiobutton(self.group_bus, text=_L["CS_BUSUSEALL"], value=1, variable=self.busMode, command=self._busModeChanged)
        self.rb_busAll.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.rb_busSel = Radiobutton(self.group_bus, text=_L["CS_BUSSELECTED"], value=2, variable=self.busMode, command=self._busModeChanged)
        self.rb_busSel.grid(row=0,column=2,padx=3,pady=3,sticky="w")
        self.entry_bussel = Entry(self.group_bus, state=DISABLED)
        self.entry_bussel.grid(row=0,column=3,padx=3,pady=3,sticky="w")
        self.rb_busRandN = Radiobutton(self.group_bus, text=_L["CS_BUSRANDOM"], value=3, variable=self.busMode, command=self._busModeChanged)
        self.rb_busRandN.grid(row=0,column=4,padx=3,pady=3,sticky="w")
        self.entry_busrandN = Entry(self.group_bus, state=DISABLED)
        self.entry_busrandN.grid(row=0,column=5,padx=3,pady=3,sticky="w")
        self.group_bus.grid(row=5,column=0,padx=3,pady=3,sticky="nesw")

        self.btn_regen = Button(self.gens, text=_L["CS_BTN_GEN"], command=self.generate)
        self.btn_regen.grid(row=6,column=0,padx=3,pady=3,sticky="w")
        self.tree.setOnSave(self.save())

        self.cslist:List[CS] = []
    
    @property
    def saved(self):
        return self.tree.saved
    
    def save(self):
        def mkFunc(s:str):
            try:
                return ConstFunc(float(s))
            except:
                return SegFunc(eval(s))
            
        def _save(data:List[tuple]):
            if not self.file: return False
            assert len(self.cslist) == len(data)
            with open(self.file, "w") as f:
                f.write(f"<?xml version='1.0' encoding='utf-8'?>\n<root>\n")
                if self.csType == FCS:
                    for i, d in enumerate(data):
                        assert len(d) == 9
                        name, slots, bus, x, y, ol, maxpc, pbuy, pcalloc = d
                        c = self.cslist[i]
                        c._name = name
                        c._slots = slots
                        c._bus = bus
                        c._x = float(x)
                        c._y = float(y)
                        c._pc_lim1 = float(maxpc) / 3600
                        if ol == ALWAYS_ONLINE: ol = "[]"
                        c._offline = RangeList(eval(ol))
                        c._pbuy = OverrideFunc(mkFunc(pbuy))
                        f.write(c.to_xml())
                        f.write("\n")
                else:
                    for i, d in enumerate(data):
                        assert len(d) == 12
                        name, slots, bus, x, y, ol, maxpc, maxpd, pbuy, psell, pcalloc, pdalloc = d
                        c = self.cslist[i]
                        c._name = name
                        c._slots = slots
                        c._bus = bus
                        c._x = float(x)
                        c._y = float(y)
                        c._pc_lim1 = float(maxpc) / 3600
                        c._pd_lim1 = float(maxpd) / 3600
                        if ol == ALWAYS_ONLINE: ol = "[]"
                        c._offline = RangeList(eval(ol))
                        c._pbuy = OverrideFunc(mkFunc(pbuy))
                        c._psell = OverrideFunc(mkFunc(psell))
                        c._pc_alloc_str = pcalloc
                        c._pd_alloc_str = pdalloc
                        f.write(c.to_xml())
                        f.write("\n")
                f.write("</root>")
            return True
        return _save

    def FindCS(self, edge:str):
        for item in self.tree.get_children():
            if self.tree.item(item, 'values')[0] == edge:
                self.tree.tree.selection_set(item)
                self.tree.tree.focus(item)
                self.tree.tree.see(item)
                break
    
    def _on_btn_find_click(self):
        self.FindCS(self.entry_find.get())
    
    def setPoly(self, val:bool):
        if not val:
            self.rb_poly.configure(state=DISABLED)
            if self.use_cscsv.get() == 2:
                self.use_cscsv.set(0)
        else:
            self.rb_poly.configure(state=NORMAL)
    
    def setCSCSV(self, val:bool):
        if not val:
            self.rb_cscsv.configure(state=DISABLED)
            if self.use_cscsv.get() == 1:
                self.use_cscsv.set(0)
        else:
            self.rb_cscsv.configure(state=NORMAL)

    def _pBuyChanged(self):
        v = self.pbuy.get()
        if v == 0:
            self.entry_pbuy.config(state=DISABLED)
        else:
            self.entry_pbuy.config(state=NORMAL)
    
    def _pSellChanged(self):
        v = self.psell.get()
        if v == 0:
            self.entry_psell.config(state=DISABLED)
        else:
            self.entry_psell.config(state=NORMAL)
    
    def _useModeChanged(self):
        v = self.useMode.get()
        if v == 0:
            self.entry_sel.config(state=DISABLED)
            self.entry_randN.config(state=DISABLED)
        elif v == 1:
            self.entry_sel.config(state=NORMAL)
            self.entry_randN.config(state=DISABLED)
        else:
            self.entry_sel.config(state=DISABLED)
            self.entry_randN.config(state=NORMAL)
    
    def _busModeChanged(self):
        v = self.busMode.get()
        if v == 0 or v == 1:
            self.entry_bussel.config(state=DISABLED)
            self.entry_busrandN.config(state=DISABLED)
        elif v == 2:
            self.entry_bussel.config(state=NORMAL)
            self.entry_busrandN.config(state=DISABLED)
        else:
            self.entry_bussel.config(state=DISABLED)
            self.entry_busrandN.config(state=NORMAL)

    @errwrapper
    def generate(self):
        seed = try_int(self.entry_seed.get(), _L["CSGUI_GEN_SEED"])
        slots = try_int(self.entry_slots.get(), _L["CSGUI_GEN_SLOTS"])
        mode = "fcs" if self.csType == FCS else "scs"

        if self.useMode.get() == 0:
            cs = ListSelection.ALL
            csCount = -1
            givenCS = []
        elif self.useMode.get() == 1:
            cs = ListSelection.GIVEN
            csCount = -1
            givenCS = try_split(self.entry_sel.get(), _L["CSGUI_GEN_GIVENCS"])
            assert not (len(givenCS) == 0 or len(givenCS) == 1 and givenCS[0] == ""), "No given CS"
        else:
            cs = ListSelection.RANDOM
            csCount = try_int(self.entry_randN.get(), _L["CSGUI_GEN_RANDOMNCS"])
            givenCS = []
        use_grid = False
        busCount = -1
        givenbus = []
        if self.busMode.get() == 0:
            use_grid = True
            bus = ListSelection.ALL
        elif self.busMode.get() == 1:
            bus = ListSelection.ALL
        elif self.busMode.get() == 2:
            bus = ListSelection.GIVEN
            givenbus = try_split(self.entry_bussel.get(), _L["CSGUI_GEN_GIVENBUS"])
            assert not (len(givenbus) == 0 or len(givenbus) == 1 and givenbus[0] == ""), "No given bus"
        else:
            bus = ListSelection.RANDOM
            busCount = try_int(self.entry_randN.get(), _L["CSGUI_GEN_RANDOMNBUS"])
        
        if self.pbuy.get() == 0:
            pbuyM = PricingMethod.RANDOM
            pbuy = 1.0
        else:
            pbuyM = PricingMethod.FIXED
            pbuy = try_float(self.entry_pbuy.get(), _L["CSGUI_GEN_PRICEBUY"])
        if self.csType == FCS:
            if self.psell.get() == 0:
                psellM = PricingMethod.RANDOM
                psell = 0
            else:
                psellM = PricingMethod.FIXED
                psell = try_float(self.entry_psell.get(), _L["CSGUI_GEN_PRICESELL"])
        else:
            psellM = PricingMethod.FIXED
            psell = 0
        self.btn_regen.config(state=DISABLED)
        self.gf(self, self.use_cscsv.get(), seed = seed, mode = mode, slots = slots,
                bus = bus, busCount = busCount, givenBus = givenbus,
                cs = cs, csCount = csCount, givenCS = givenCS, 
                priceBuyMethod = pbuyM, priceBuy = pbuy, priceSellMethod = psellM, 
                priceSell = psell, hasSell = self.csType == SCS, use_grid = use_grid)
    
    def load(self, file:str):
        self._Q.submit("loaded", self.__load, file)
            
    def clear(self):
        self.tree.clear()
        self.lb_cnt.config(text=_L["LB_COUNT"].format(0))
    
    def __load(self, file:str):
        try:
            self.cslist = LoadCSList(file, self.csType)
        except Exception as e:
            showerr(_L["ERROR_LOADING"].format(file, e))
            return
        self.file = file
        self.tree.clear()
        self.lb_cnt.config(text=_L["LB_COUNT"].format(len(self.cslist)))
        if self.csType == FCS:
            for cs in self.cslist:
                ol = str(cs._offline) if len(cs._offline)>0 else ALWAYS_ONLINE
                v = (cs.name, cs.slots, cs.bus, cs._x, cs._y, ol, cs._pc_lim1 * 3600, cs.pbuy, cs._pc_alloc_str)
                self._Q.delegate(self.tree.insert, "", "end", values=v)
        else:
            for cs in self.cslist:
                assert isinstance(cs, SCS)
                ol = str(cs._offline) if len(cs._offline)>0 else ALWAYS_ONLINE
                v = (cs.name, cs.slots, cs.bus, cs._x, cs._y, ol, cs._pc_lim1 * 3600, cs._pd_lim1 * 3600, 
                            cs.pbuy, cs._psell, cs._pc_alloc_str, cs._pd_alloc_str)
                self._Q.delegate(self.tree.insert, "", "end", values=v)


from v2simux_gui.com_no_vx import *

_ = LangLib(["zh_CN", "en_US"])
_.SetLangLib("zh_CN",
    MB_INFO = "信息",
    MENU_LANG = "语言",
    MENU_LANG_AUTO = "(自动检测)",
    LANG_RESTART = "语言已更改，请重启程序以应用更改。",
)

_.SetLangLib("en_US",
    MB_INFO = "Information",
    MENU_LANG = "Language",
    MENU_LANG_AUTO = "(Auto Detect)",
    LANG_RESTART = "Language has been changed. Please restart the application to apply the changes.",
)

def setLang(lang_code:str):
    title = _["MB_INFO"]
    info = _["LANG_RESTART"]
    def _f():
        nonlocal lang_code, title, info
        LangConfig.SetLangCode(lang_code)
        MB.showinfo(title, info)
    return _f

def add_lang_menu(parent: Menu):
    menuLang = Menu(parent, tearoff=False)
    parent.add_cascade(label=_["MENU_LANG"], menu=menuLang)
    menuLang.add_command(label=_["MENU_LANG_AUTO"], command=setLang("<auto>"))
    menuLang.add_command(label="English (United States)", command=setLang("en_US"))
    menuLang.add_command(label="中文 (简体)", command=setLang("zh_CN"))

__all__ = ["add_lang_menu"]



from v2simux_gui.common import *

import functools

_L = LangLib.Load(__file__)

V2SIM_UX_DIR = Path(__file__).parent.parent.parent
EXT_COMP = str(V2SIM_UX_DIR / "external_components")
SIM_YES = "YES"
SIM_NO = "NO"

def showerr(msg:str):
    MB.showerror(_L["MB_ERROR"], msg)

def showwarn(msg:str):
    MB.showwarning(_L["MB_INFO"], msg)

def try_int(s:str, name:str) -> int:
    try: return int(s)
    except: raise ValueError(f"Invalid {name}. Must be an integer.")

def try_float(s:str, name:str) -> float:
    try: return float(s)
    except: raise ValueError(f"Invalid {name}. Must be a float.")

def try_split(s:str, name:str, sep:str=',') -> List[str]:
    try:
        parts = s.split(sep)
        parts = [p.strip() for p in parts if p.strip()]
        return parts
    except:
        raise ValueError(f"Invalid {name}. Must be a list of strings separated by '{sep}'.")

def errwrapper(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            showerr(str(e))
    return wrapped

__all__ = ["showerr", "showwarn", "try_int", "try_float", "try_split", "errwrapper", "EXT_COMP", "SIM_YES", "SIM_NO", "V2SIM_UX_DIR"]


from ..plugins import *
from ..statistics import *
from typing import Any, List
from feasytools import SegFunc
import math
    
class StatisticsNotSupportedError(Exception): pass

def _parse_val(x:str)->List[Any]:
    v = []
    v0 = 0
    for c in x:
        if c.isdigit():
            v0 = v0*10 + int(c)
        else:
            if v0>0:
                v.append(v0)
                v0 = 0
            v.append(ord(c))
    if v0>0: v.append(v0)
    return v

TO_BE_LOADED = True

class ReadOnlyStatistics(StaReader):
    def has_FCS(self)->bool: return self.__fcs_head is not None
    def has_SCS(self)->bool: return self.__scs_head is not None
    def has_EV(self)->bool: return FILE_EV in self
    def has_GEN(self)->bool: return FILE_GEN in self
    def has_BUS(self)->bool: return FILE_BUS in self
    def has_LINE(self)->bool: return FILE_LINE in self
    def has_PVW(self)->bool: return FILE_PVW in self
    def has_ESS(self)->bool: return FILE_ESS in self
    
    def __loadhead(self, name:str) -> List[str]:
        def __trans(x:str):
            p = x.rfind("#")
            if p>=0: return x[:p]
            return x
        return list(set(__trans(x) for x in self.GetTable(name).keys()))
    
    def __init__(self, path:str):
        super().__init__(path)
        self.root = path
        self.__fcs_head = None if FILE_FCS not in self else TO_BE_LOADED
        self.__scs_head = None if FILE_SCS not in self else TO_BE_LOADED
        self.__ev_head  = None if FILE_EV  not in self else TO_BE_LOADED
        self.__gen_head = None if FILE_GEN not in self else TO_BE_LOADED
        self.__bus_head = None if FILE_BUS not in self else TO_BE_LOADED
        self.__line_head = None if FILE_LINE not in self else TO_BE_LOADED
        self.__pvw_head = None if FILE_PVW not in self else TO_BE_LOADED
        self.__ess_head = None if FILE_ESS not in self else TO_BE_LOADED
        
    @property
    def FCS_head(self)->List[str]: 
        assert self.__fcs_head is not None, "CS properties not supported"
        if not isinstance(self.__fcs_head, list):
            self.__fcs_head = self.__loadhead(FILE_FCS)
            self.__fcs_head.sort(key=_parse_val)
        return self.__fcs_head

    @property
    def SCS_head(self)->List[str]: 
        assert self.__scs_head is not None, "CS properties not supported"
        if not isinstance(self.__scs_head, list):
            self.__scs_head = self.__loadhead(FILE_SCS)
            self.__scs_head.sort(key=_parse_val)
        return self.__scs_head
    
    @property
    def veh_head(self)->List[str]:
        assert self.__ev_head is not None, "Vehicle properties not supported"
        if not isinstance(self.__ev_head, list):
            self.__ev_head = self.__loadhead(FILE_EV)
            self.__ev_head.sort(key=_parse_val)
        return self.__ev_head
    
    @property
    def gen_head(self)->List[str]:
        assert self.__gen_head is not None, "Generator properties not supported"
        if not isinstance(self.__gen_head, list):
            self.__gen_head = self.__loadhead(FILE_GEN)
            for itm in GEN_TOT_ATTRIB:
                if itm in self.__gen_head:
                    self.__gen_head.remove(itm)
            self.__gen_head.sort(key=_parse_val)
        return self.__gen_head
    
    @property
    def bus_head(self)->List[str]:
        assert self.__bus_head is not None, "Bus properties not supported"
        if not isinstance(self.__bus_head, list):
            self.__bus_head = self.__loadhead(FILE_BUS)
            for itm in BUS_TOT_ATTRIB:
                if itm in self.__bus_head:
                    self.__bus_head.remove(itm)
            self.__bus_head.sort(key=_parse_val)
        return self.__bus_head
    
    @property
    def line_head(self)->List[str]:
        assert self.__line_head is not None, "Line properties not supported"
        if not isinstance(self.__line_head, list):
            self.__line_head = self.__loadhead(FILE_LINE)
            for itm in LINE_ATTRIB:
                if itm in self.__line_head:
                    self.__line_head.remove(itm)
            self.__line_head.sort(key=_parse_val)
        return self.__line_head
    
    @property
    def pvw_head(self)->List[str]:
        assert self.__pvw_head is not None, "PV & Wind properties not supported"
        if not isinstance(self.__pvw_head, list):
            self.__pvw_head = self.__loadhead(FILE_PVW)
            for itm in PVW_ATTRIB:
                if itm in self.__pvw_head:
                    self.__pvw_head.remove(itm)
            self.__pvw_head.sort(key=_parse_val)
        return self.__pvw_head

    @property
    def ess_head(self)->List[str]:
        assert self.__ess_head is not None, "ESS properties not supported"
        if not isinstance(self.__ess_head, list):
            self.__ess_head = self.__loadhead(FILE_ESS)
            for itm in ESS_ATTRIB:
                if itm in self.__ess_head:
                    self.__ess_head.remove(itm)
            self.__ess_head.sort(key=_parse_val)
        return self.__ess_head

    def FCS_attrib_of(self,cs:str, attrib:str)->SegFunc: 
        '''Charging station information'''
        assert attrib in CS_ATTRIB, f"Invalid CS property: {attrib}"
        if cs == "<sum>":
            d = [self.GetColumn(FILE_FCS,c+"#"+attrib) for c in self.FCS_head]
            return SegFunc.qs(d)
        return self.GetColumn(FILE_FCS,cs+"#"+attrib)
    
    def SCS_attrib_of(self,cs:str,attrib:str)->SegFunc: 
        '''Charging station information'''
        assert attrib in CS_ATTRIB, f"Invalid CS property: {attrib}"
        if cs == "<sum>":
            d = [self.GetColumn(FILE_SCS,c+"#"+attrib) for c in self.SCS_head]
            return SegFunc.qs(d)
        return self.GetColumn(FILE_SCS,cs+"#"+attrib)
    
    def FCS_load_of(self,cs:str)->SegFunc:
        '''Charging power'''
        return self.FCS_attrib_of(cs,"c")
    
    def FCS_load_all(self,tl=-math.inf,tr=math.inf)->List[SegFunc]:
        '''Charging power of all CS'''
        return [self.FCS_load_of(cs).slice(tl,tr) for cs in self.FCS_head]
    
    def FCS_count_of(self,cs:str)->SegFunc:
        '''Number of vehicles in the CS'''
        return self.FCS_attrib_of(cs,"cnt")
    
    def FCS_pricebuy_of(self,cs:str)->SegFunc:
        '''Buy price'''
        return self.FCS_attrib_of(cs,"pb")
    
    def SCS_charge_load_of(self,cs:str)->SegFunc:
        '''Charging power'''
        return self.SCS_attrib_of(cs,"c")
    
    def SCS_charge_load_all(self,tl=-math.inf,tr=math.inf)->List[SegFunc]:
        '''Charging power of all CS'''
        return [self.SCS_charge_load_of(cs).slice(tl,tr) for cs in self.SCS_head]
    
    def SCS_v2g_load_of(self,cs:str)->SegFunc:
        '''Discharging power (V2G)'''
        return self.SCS_attrib_of(cs,"d")
    
    def SCS_v2g_load_all(self,tl=-math.inf,tr=math.inf)->List[SegFunc]:
        '''Discharging power (V2G) of all CS'''
        return [self.SCS_v2g_load_of(cs).slice(tl,tr) for cs in self.SCS_head]
    
    def SCS_v2g_cap_of(self,cs:str)->SegFunc:
        '''V2G capacity'''
        return self.SCS_attrib_of(cs,"v2g")
    
    def SCS_v2g_cap_all(self,tl=-math.inf,tr=math.inf)->List[SegFunc]:
        '''V2G capacity of all CS'''
        return [self.SCS_v2g_cap_of(cs).slice(tl,tr) for cs in self.SCS_head]
    
    def SCS_net_load_of(self,cs:str)->SegFunc:
        '''Net charging power'''
        return self.SCS_charge_load_of(cs) - self.SCS_v2g_load_of(cs)
    
    def SCS_net_load_all(self,tl=-math.inf,tr=math.inf)->List[SegFunc]:
        '''Net charging power of all CS'''
        return [self.SCS_net_load_of(cs).slice(tl,tr) for cs in self.SCS_head]
    
    def SCS_count_of(self,cs:str)->SegFunc:
        '''Number of vehicles in the CS'''
        return self.SCS_attrib_of(cs,"cnt")
    
    def SCS_pricebuy_of(self,cs:str)->SegFunc:
        '''Buy price'''
        return self.SCS_attrib_of(cs,"pb")
    
    def SCS_pricesell_of(self,cs:str)->SegFunc:
        '''Sell price'''
        return self.SCS_attrib_of(cs,"ps")
    
    def EV_attrib_of(self,veh:str,attrib:str)->SegFunc: 
        '''EV information'''
        assert attrib in EV_ATTRIB, f"Invalid EV property: {attrib}"
        return self.GetColumn(FILE_EV,veh+"#"+attrib)
    
    def EV_net_cost_of(self,veh:str)->SegFunc: 
        '''EV net cost'''
        return self.GetColumn(FILE_EV,veh+"#cost")-self.GetColumn(FILE_EV,veh+"#earn")
    
    def G_attrib_of(self,g:str,attrib:str)->SegFunc: 
        '''Generator information'''
        assert attrib in GEN_ATTRIB
        return self.GetColumn(FILE_GEN,g+"#"+attrib)
    
    def G_total(self,attrib:str)->SegFunc:
        '''Total generation data'''
        assert attrib in GEN_TOT_ATTRIB
        return self.GetColumn(FILE_GEN,attrib)
    
    def bus_attrib_of(self,b:str,attrib:str)->SegFunc: 
        '''Bus information'''
        assert attrib in BUS_ATTRIB
        return self.GetColumn(FILE_BUS,b+"#"+attrib)
    
    def bus_total(self,attrib:str)->SegFunc:
        '''Total bus data'''
        assert attrib in BUS_TOT_ATTRIB
        return self.GetColumn(FILE_BUS,attrib)
    
    def line_attrib_of(self,l:str,attrib:str)->SegFunc: 
        '''Line information'''
        assert attrib in LINE_ATTRIB
        return self.GetColumn(FILE_LINE,l+"#"+attrib)
    
    def pvw_attrib_of(self,p:str,attrib:str)->SegFunc:
        '''PV and wind information'''
        assert attrib in PVW_ATTRIB
        return self.GetColumn(FILE_PVW,p+"#"+attrib)
    
    def ess_attrib_of(self,e:str,attrib:str)->SegFunc:
        '''ESS information'''
        assert attrib in ESS_ATTRIB
        return self.GetColumn(FILE_ESS,e+"#"+attrib)


from v2simux_gui.common import *

from .scrtv import ScrollableTreeView


_L = LangLib.Load(__file__)


class PropertyPanel(Frame):
    def __onclick(self, event):
        if len(self.tree.selection()) == 0:
            self.__desc_var.set(_L["PROP_NODESC"])
            return
        self.selected_item = self.tree.selection()[0]
        selected_row = self.tree.item(self.selected_item, "values")[0]
        self.__desc_var.set(self.__em.get_desc(selected_row))

    def setData(self, data:Dict[str, Any], edit_modes:ConfigItemDict):
        self.tree.tree_item_edit_done(None)
        self.data = data
        self.__em:ConfigItemDict = edit_modes
        if edit_modes is None: edit_modes = {}
        self.tree.clear()
        for l, r in data.items():
            self.tree.insert("", "end", values=(l, r))
            self.tree.setCellEditMode(l, "d", edit_modes.get(l))

    def setDataEmpty(self):
        self.setData({}, ConfigItemDict())
        
    def setData2(self, *val_and_modes:Tuple[Any, ConfigItem]):
        data = {}
        edit_modes = ConfigItemDict()
        for v, m in val_and_modes:
            data[m.name] = v
            edit_modes[m.name] = m
        self.setData(data, edit_modes)
    
    def __init__(self, master, data:Dict[str,str], edit_modes:ConfigItemDict, **kwargs):
        super().__init__(master, **kwargs)
        self.tree = ScrollableTreeView(self, allowSave=False)
        self.tree['show'] = 'headings'
        self.tree["columns"] = ("t", "d")
        self.tree.column("t", width=120, stretch=NO)
        self.tree.column("d", width=120, stretch=YES)
        self.tree.heading("t", text=_L["PROPERTY"])
        self.tree.heading("d", text=_L["VALUE"])
        self.tree.tree.bind("<<TreeviewSelect>>", self.__onclick)
        self.tree.pack(fill="both", expand=True)
        self.__desc_var = StringVar(self, _L["PROP_NODESC"])
        self.__desc = Label(self, textvariable=self.__desc_var)
        self.__desc.pack(fill="x", expand=False)
        self.setData(data, edit_modes)

    def getAllData(self) -> Dict[str, str]:
        res:Dict[str, str] = {}
        for i in self.tree.get_children():
            x = self.tree.tree.item(i, "values")
            res[x[0]] = x[1]
        return res

class PropertyEditor(Toplevel):
    def __init__(self, data:Dict[str,str], var:StringVar, edit_modes:ConfigItemDict):
        super().__init__()
        self.title(_L["PROPERTY_EDITOR"])
        self.__panel = PropertyPanel(self, data, edit_modes)
        self.__panel.pack(fill="both", expand=True)
        self.__fr = Frame(self)
        self.__fr.pack(fill="x", expand=False)
        self.__btn_save = Button(self.__fr, text=_L["SAVE_AND_CLOSE"], command=self.save)
        self.__btn_save.grid(row=0,column=4,padx=3,pady=3,sticky="e")
        self.var = var
    
    def getAllData(self) -> Dict[str,str]:
        return self.__panel.getAllData()
    
    def save(self):
        d = self.getAllData()
        self.var.set(repr(d))
        self.destroy()

__all__ = ["PropertyPanel", "PropertyEditor"]


import os
from collections import defaultdict
from pathlib import Path
from feasytools import SegFunc
from typing import Type, Optional, Union
from ..plugins import *
from .logcs import *
from .logev import *
from .loggr import *
from ..locale import Lang


class StaPool:
    def __init__(self, load_internal_logger: bool = True):
        self.__ava_logger: Dict[str, type] = {}
        if load_internal_logger:
            self.__ava_logger = {
                FILE_FCS: StaFCS,
                FILE_SCS: StaSCS,
                FILE_EV: StaEV,
                FILE_GEN: StaGen,
                FILE_BUS: StaBus,
                FILE_LINE: StaLine,
                FILE_PVW: StaPVWind,
                FILE_ESS: StaESS,
            }

    def Register(self, name: str, base: Type) -> None:
        """Register a statistic item"""
        if name in self.__ava_logger:
            raise ValueError(Lang.ERROR_STA_REGISTERED.format(name))
        assert issubclass(base, StaBase)
        self.__ava_logger[name] = base

    def Get(self, name: str) -> Type:
        """Get a statistic item"""
        return self.__ava_logger[name]

    def GetAllLogItem(self) -> List[str]:
        """Get all statistic items"""
        return list(self.__ava_logger.keys())
    
    def GetAllLogItemLocalizedName(self) -> List[str]:
        """Get all statistic items with localized names"""
        return [sta.GetLocalizedName() for sta in self.__ava_logger.values()]


class StaWriter:
    """Statistic data recorder"""
    __items: Dict[str, StaBase]

    def Writer(self, name: str):
        return self.__items[name].Writer

    def __init__(
        self,
        path: Union[str, Path],
        tinst: TrafficInst,
        plugins: Dict[str, PluginBase],
        staPool: StaPool,
        items: Optional[List[str]] = None,
    ):
        """
        Initialize
            path: Path to the output file
            tinst: Traffic instance
            plugins: Loaded plugins
            staPool: Statistics items' pool
            items: List of statistic items to record. Can be added later by 'Add' function. If None, no item is added now.
        """
        self.__path = path if isinstance(path, str) else str(path)
        self.__items = {}
        self.__inst = tinst
        self.__plug = plugins
        self.__pool = staPool

        if items is not None:
            for itm in items:
                self.Add(itm)

    def Add(self, sta_name: str) -> None:
        """Add a statistic item, select from the registered items of StaMan"""
        sta_type = self.__pool.Get(sta_name)
        if sta_name in self.__items:
            raise ValueError(Lang.ERROR_STA_ADDED.format(sta_name))
        self.__items[sta_name] = sta_type(self.__path, self.__inst, self.__plug)

    def Log(self, time: int):
        for item in self.__items.values():
            try:
                item.LogOnce()
            except Exception as e:
                print(Lang.ERROR_STA_LOG_ITEM.format(item._name, e))
                raise e

    def close(self):
        for item in self.__items.values():
            try:
                item.close()
            except Exception as e:
                print(Lang.ERROR_STA_CLOSE_ITEM.format(item._name, e))
                raise e

class _CSVTable:
    def force_load(self):
        lastTime:Dict[str,int] = defaultdict(lambda:-1)
        data = self.__f.readlines()
        lt = -1
        for i, line in enumerate(data,2):
            time, item, value = line.strip().split(",")
            if self._mp is not None: item = self._mp[item]
            time = int(time) if time != "" else lt
            assert time > lastTime[item], f"Item {item} @ line {i}: Time must be increasing, but value to add ({time}) is smaller or equal to the last time ({lastTime[item]})"
            lastTime[item] = time
            lt = time
            self.__data[item].add(time, float(value))
        if self.__head is None: self.__head = list(self.__data.keys())
        self.__lt = lt
        self.__f.close()
        self.__loaded = True

    def __init__(self, filename:str, preload:bool=False):
        self.__data:Dict[str, SegFunc] = defaultdict(SegFunc)
        self.__loaded = False
        self.__f = open(filename, "r")
        head = self.__f.readline().strip()
        self._mp = None
        if head == "C":
            head = self.__f.readline().strip().split(",")
            self.__head = head
            self._mp = {to_base62(i):item for i, item in enumerate(head)}
            head = self.__f.readline().strip()
        else:
            self.__head = None
        header = head.split(",")
        assert len(header) == 3 and header[0] == "Time" and header[1] == "Item" and header[2] == "Value"
        self.__lt = -1
        if preload: self.force_load()
    
    def __getitem__(self, key:str)->SegFunc:
        if not self.__loaded: self.force_load()
        return self.__data[key]
    
    def __contains__(self, key:str)->bool:
        if not self.__loaded: self.force_load()
        return key in self.__data
    
    def keys(self) -> List[str]:
        if self.__head is None: self.force_load()
        return self.__head # type: ignore
    
    @property
    def LastTime(self)->int:
        if not self.__loaded: self.force_load()
        return self.__lt
    
class StaReader:
    """
    Generic purpose readonly statistics reader, created from a folder.
    If you are reading results from a standard V2Sim simulation, using ReadOnlyStatistics is better than using this class, for it supports detailed API.
    """
    def __init__(
        self,
        path: str,
        sta_pool: Optional[StaPool] = None,
    ):
        """
        Initialize
            path: Path to the results folder
            sta_pool: Statistics items' pool, for checking whether an item exists. None for not checking.
        """
        work_dir = Path(path)
        dir_con = os.listdir(path)
        self.__items: Dict[str, _CSVTable] = {}
        for file in dir_con:
            if file.endswith(".csv"):
                fname = file[:-4]  # Remove .csv suffix
                if sta_pool is None or sta_pool.Get(fname) is not None:
                    self.__items[fname] = _CSVTable(str(work_dir / file))
            else:
                continue

    def __contains__(self, table_name: str) -> bool:
        return table_name in self.__items

    def __getitem__(self, table_name: str) -> _CSVTable:
        return self.__items[table_name]
    
    def GetColumn(self, table_name:str, item:str)->SegFunc:
        if table_name not in self.__items:
            raise ValueError(f"Table '{table_name}' not found")
        if item not in self.__items[table_name]:
            return SegFunc([(0, 0.0)])
            #raise ValueError(f"Item '{item}' not found in table '{table_name}'")
        return self.__items[table_name][item]
    
    def GetTable(self, table_name:str)->_CSVTable:
        if table_name not in self.__items:
            raise ValueError(f"Table '{table_name}' not found")
        return self.__items[table_name]
    
    @property
    def LastTime(self)->int:
        t = 0
        for table in self.__items.values():
            t = max(t, table.LastTime)
        return t


from .base import *
from .pdn import PluginPDN
from .v2g import PluginV2G
from .ocur import PluginOvercurrent

_internal_plugins = {
    "pdn": (PluginPDN,[]),
    "v2g": (PluginV2G,["pdn"]),
    "ocur": (PluginOvercurrent,["pdn"]),
}

class PluginError(Exception):
    pass

class PluginPool:
    '''Plugin pool'''
    def __init__(self, use_internal_plugins:bool = True):
        '''
        Initialize
            use_internal_plugins: Whether to load internal plugins
        '''
        self.__curPlugins:Dict[str,Tuple[type,List[str]]] = {}
        if use_internal_plugins:
            for k,(p,d) in _internal_plugins.items():
                self._Register(k,p,d)
        
        for key,(_,deps) in self.__curPlugins.items():
            for d in deps:
                if d not in self.__curPlugins:
                    print(self.__curPlugins)
                    raise PluginError(Lang.PLG_DEPS_NOT_REGISTERED.format(key,d))

    def __getitem__(self,name:str)->Tuple[type,List[str]]:
        '''Get plugin by name'''
        return self.__curPlugins[name]
    
    def GetPluginType(self,name:str)->type:
        '''Get plugin type by name'''
        return self.__curPlugins[name][0]
    
    def GetPluginDependencies(self,name:str)->List[str]:
        '''Get plugin dependencies by name'''
        return self.__curPlugins[name][1]
    
    def GetAllPlugins(self,)->Dict[str,Tuple[type,List[str]]]:
        '''Get all plugins'''
        return self.__curPlugins
    
    def GetAllPluginNames(self)->List[str]:
        '''Get all plugin names'''
        return list(self.__curPlugins.keys())
    
    def __contains__(self,name:str)->bool:
        '''Check if plugin exists'''
        return self.__curPlugins.__contains__(name)
    
    def _Register(self,name:str,plugin:type,deps:List[str]):
        '''Register new plugin to plugin pool without checking dependencies'''
        if name in self.__curPlugins:
            raise PluginError(Lang.PLG_REGISTERED.format(name))
        for d in deps:
            if not isinstance(d,str):
                raise PluginError(Lang.PLG_DEPS_MUST_BE_STRLIST)
        self.__curPlugins[name] = (plugin, deps)
    
    def Register(self,name:str,plugin:type,deps:List[str]):
        '''Register new plugin to plugin pool'''
        if not issubclass(plugin,PluginBase):
            raise PluginError(Lang.PLG_NOT_SUBCLASS.format(plugin))
        for d in deps:
            if d not in self.__curPlugins:
                raise PluginError(Lang.PLG_DEPS_NOT_REGISTERED.format(name,d))
        self._Register(name,plugin,deps)

class PluginMan:
    def __init__(self, plg_xml:Optional[str], res_dir:Path, inst:TrafficInst, no_plg:List[str], plugin_pool:PluginPool, initial_state:Optional[Dict[str, object]] = None):
        '''
        Load plugins from file
            plg_xml: Plugin configuration file path, None means not load
            res_dir: Result directory
            inst: Traffic simulation instance
            no_plg: Plugins not to load
            plugin_pool: Available plugin pool
            initial_state_file: Initial plugin states, if specified, load plugin states from it
        '''
        self.__curPlugins:Dict[str, PluginBase] = {}
        if plg_xml is None:
            return
        if Path(plg_xml).exists() == False:
            print(Lang.PLG_NOT_EXIST.format(plg_xml))
            return
        root = ET.ElementTree(file=plg_xml).getroot()
        work_dir = Path(plg_xml).parent
        if root is None:
            raise PluginError(Lang.PLG_NOT_EXIST_OR_BAD.format(plg_xml))
        
        # Load plugins
        for itm in root:
            if itm.tag in no_plg or itm.attrib.get("enabled") == "NO":
                continue
            if itm.tag not in plugin_pool:
                raise PluginError(Lang.PLG_INVALID_PLUGIN.format(itm.tag))
            plugin_type, dependencies = plugin_pool[itm.tag]
            deps:List[PluginBase] = []
            for d in dependencies:
                if d not in self.__curPlugins:
                    raise PluginError(Lang.PLG_DEPS_NOT_LOADED.format(itm.tag,d))
                deps.append(self.__curPlugins[d])
            self.Add(plugin_type(inst, itm, work_dir, res_dir, plg_deps=deps,
                initial_state=(initial_state.pop(itm.tag) if initial_state else None)))

        if initial_state and len(initial_state) > 0:
            raise PluginError("Some plugins in the initial state file are not loaded: " + ",".join(initial_state.keys()))
    
    def PreSimulationAll(self):
        '''Execute all plugins PreSimulation, return all plugins return value'''
        for p in self.__curPlugins.values():
            p._presim()
    
    def PostSimulationAll(self):
        '''Execute all plugins PostSimulation, return all plugins return value'''
        for p in self.__curPlugins.values():
            p._postsim()

    def PreStepAll(self, _t:int) -> Dict[str,object]:
        '''Execute all plugins PreStep, return all plugins return value'''
        ret:Dict[str,object] = {}
        for k,p in self.__curPlugins.items():
            ret[k] = p._precall(_t)
        return ret
    
    def PostStepAll(self, _t:int) -> Dict[str,object]:
        '''Execute all plugins PreStep, return all plugins return value'''
        ret:Dict[str,object] = {}
        for k,p in self.__curPlugins.items():
            ret[k] = p._postcall(_t)
        return ret
    
    def Add(self, plugin:PluginBase):
        '''Add new plugin to current plugin list'''
        if plugin.Name in self.__curPlugins:
            raise PluginError(Lang.PLG_ALREADY_EXISTS.format(plugin.Name))
        self.__curPlugins[plugin.Name] = plugin

    def GetPluginByName(self, name:str) -> PluginBase:
        '''Get plugin by name'''
        return self.__curPlugins[name]
    
    def GetPlugins(self) -> Dict[str,PluginBase]:
        '''Get all plugins'''
        return self.__curPlugins

    def SaveStates(self) -> Dict[str, object]:
        '''Save all plugin states'''
        ret = {}
        for name, p in self.__curPlugins.items():
            ret[name] = p._save_state()
        return ret


from .base import StaBase, cross_list
from .manager import StaPool, StaReader, StaWriter
from .logcs import StaFCS, StaSCS, FILE_FCS, FILE_SCS, CS_ATTRIB
from .logev import StaEV, FILE_EV, EV_ATTRIB
from .loggr import (
    StaGen, StaBus, StaLine, StaESS, StaPVWind,
    FILE_GEN, FILE_BUS, FILE_LINE, FILE_ESS, FILE_PVW,
    GEN_ATTRIB, BUS_ATTRIB, LINE_ATTRIB, ESS_ATTRIB, PVW_ATTRIB,
    BUS_TOT_ATTRIB, GEN_TOT_ATTRIB, 
)


import gzip
import dill as pickle
import importlib, os, queue, shutil, signal, time, sys
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union
from feasytools import ArgChecker, time2str
from pathlib import Path
from .plotkit import AdvancedPlot
from .plugins import *
from .statistics import *
from .traffic import *
from .locale import Lang
from .trafficgen import TrafficGenerator
import inspect

PLUGINS_FILE = "plugins.gz"
RESULTS_FOLDER = "results"
TRIP_EVENT_LOG = "cproc.clog"
SIM_INFO_LOG = "cproc.log"


@dataclass
class MsgPack:
    clntID:int
    cmd:str
    obj:Any = None


def load_external_components(
    external_plugin_dir: Union[str, Path], plugin_pool: PluginPool, sta_pool: StaPool
):
    exp = Path(external_plugin_dir).absolute() if isinstance(external_plugin_dir, str) else external_plugin_dir.absolute()
    if not (exp.exists() and exp.is_dir()):
        return
    sys.path.append(str(exp))
    for module_file in exp.iterdir():
        if not (
            module_file.is_file()
            and module_file.suffix == ".py"
            and not module_file.name.startswith("_")
        ):
            continue
        module_name = module_file.stem
        try:
            module = importlib.import_module(module_name)
        except Exception as e:
            print(Lang.WARN_EXT_LOAD_FAILED.format(module_name, e))
            module = None
        if hasattr(module, "plugin_exports"):
            try:
                plugin_pool._Register(*module.plugin_exports) # type: ignore
            except Exception as e:
                print(Lang.WARN_EXT_INVALID_PLUGIN.format(module_name, e))
        if hasattr(module, "sta_exports"):
            try:
                sta_pool.Register(*module.sta_exports) # type: ignore
            except Exception as e:
                print(Lang.WARN_EXT_INVALID_STA.format(module_name, e))
                
def get_sim_params(
        args:Union[str, ArgChecker],
        plg_pool:PluginPool,
        sta_pool:StaPool,
        check_illegal:bool = True
    )->Dict[str,Any]:
    '''
    Get simulation parameters used by the simulate function
        args: Command line parameters or ArgChecker instance
        plg_pool: Plugin pool
        sta_pool: Statistical item pool
        check_illegal: Whether to check illegal parameters, default is True
    '''
    if isinstance(args, str):
        args = ArgChecker(args, force_parametric=["plot", "gen-veh", "gen-fcs", "gen-scs"])
    if isinstance(args, ArgChecker):
        kwargs = {
            "cfgdir":               args.pop_str("d"),
            "outdir":               args.pop_str("o", ""),
            "outdir_direct":        args.pop_str("od", ""),
            "traffic_step":         args.pop_int("l", 10),
            "start_time":           args.pop_int("b", 0),
            "break_at":             args.pop_int("break-at", -1),
            "end_time":             args.pop_int("e", 172800),
            "no_plg":               args.pop_str("no-plg", ""),
            "log":                  args.pop_str("log", "fcs,scs"),
            "seed":                 args.pop_int("seed", time.time_ns() % 65536),
            "copy":                 args.pop_bool("copy"),
            "gen_veh_command":      args.pop_str("gen-veh", ""),
            "gen_fcs_command":      args.pop_str("gen-fcs", ""),
            "gen_scs_command":      args.pop_str("gen-scs", ""),
            "plot_command":         args.pop_str("plot", ""),
            "initial_state":        args.pop_str("initial-state", ""),
            "load_last_state":      args.pop_bool("load-last-state"),
            "save_on_abort":        args.pop_bool("save-on-abort"),
            "save_on_finish":       args.pop_bool("save-on-finish"),
            "copy_state":           args.pop_bool("copy-state"),
            "route_algo":           args.pop_str("route-algo", "astar"),
            "show_uxsim_info":      args.pop_bool("show-uxsim-info"),
            "randomize_uxsim":      args.pop_bool("randomize-uxsim"),
            "no_parallel":          args.pop_bool("no-parallel"),
            "plg_pool":             plg_pool,
            "sta_pool":             sta_pool,
        }
    if check_illegal and len(args) > 0:
        for key in args.keys():
            raise ValueError(Lang.ERROR_ILLEGAL_CMD.format(key))
    return kwargs

def _calc_output_folder(cfgdir: str, outdir: str, outdir_direct: str) -> Path:
    if outdir_direct != "":
        pres = Path(outdir_direct)
    else:
        if outdir == "":
            pres = Path(cfgdir) / RESULTS_FOLDER
        else:
            pres = Path(outdir) / Path(cfgdir).name
    return pres

class V2SimInstance:
    def __mpsend(self, con:str, obj:Any = None):
        if self.__mpQ:
            try:
                self.__mpQ.put_nowait(MsgPack(self.__clntID, con, obj))
            except:
                print(Lang.WARN_SIM_COMM_FAILED)
    
    def __print(self, con:str="", *, file:Any = sys.stdout, end="\n"):
        if not self.__silent:
            print(con, file=file, end=end)

    def __init__(
        self, 
        cfgdir: str,
        outdir: str = "",
        *,
        outdir_direct: str = "",
        plg_pool: Optional[PluginPool] = None,
        sta_pool: Optional[StaPool] = None,
        gen_veh_command:str = "", 
        gen_fcs_command:str = "", 
        gen_scs_command:str = "", 
        alt_command:Optional[Dict[str,str]] = None,
        plot_command:str = "",
        traffic_step: int = 10,      
        start_time: int = 0,        
        end_time: int = 172800,
        break_at: int = -1,
        no_plg: str = "",            
        log: str = "fcs, scs",               
        seed: int = 0,              
        copy: bool = False,
        vb = None,                
        silent: bool = False,           
        mpQ: Optional[queue.Queue] = None, 
        clntID: int = -1,
        initial_state: str = "",
        load_last_state: bool = False,
        save_on_abort: bool = False,
        save_on_finish: bool = False,
        copy_state: bool = False,
        route_algo: str = "astar",
        show_uxsim_info: bool = False,
        randomize_uxsim: bool = False,
        no_parallel: bool = False,
        _unsafe_traffic_state_to_load:Optional[Dict[str,Any]] = None,
        _unsafe_inst_state_to_load:Optional[Dict[str,Any]] = None,
        _unsafe_plugin_state_to_load:Optional[Dict[str,Any]] = None,
    ):
        '''
        Initialization
            cfgdir: Configuration folder
            outdir: Output folder. Actual results will be saved in a subfolder named by the configuration folder
            outdir_direct: Direct output folder
            plg_pool: Available plugin pool
            sta_pool: Available statistical item pool
            gen_veh_command: command to generate vehicle
            gen_fcs_command: Generate fast charging station command
            gen_scs_command: Generate slow charging station command
            plot_command: Plot command
            traffic_step: Simulation step
            start_time: Start time
            end_time: End time
            no_plg: Disabled plugins, separated by commas
            log: Data to be recorded, separated by commas
            seed: Randomization seed
            copy: Whether to copy the configuration file after the simulation ends
            vb: Whether to enable the visualization window, None means not enabled, when running this function in multiple processes, please set to None
            silent: Whether to silent mode, default is False, when running this function in multiple processes, please set to True
            mpQ: Queue for communication with the main process when running this function in multiple processes, set to None if not using multi-process function
            clntID: Identifier of this process when running this function in multiple processes, set to -1 if not using multi-process function
            initial_state: Folder of the initial state of the simulation
            load_last_state: Load the state in result dir if there is a state folder. This will override initial_state if there is a state folder.
            save_on_abort: Whether to save the state when Ctrl+C is pressed
            save_on_finish: Whether to save the state when the simulation ends
            copy_state: Whether to copy the state folder after the simulation ends or when Ctrl+C is pressed
            route_algo: Routing algorithm, can be dijkstra or astar
            show_uxsim_info: Whether to show the information of uxsim
            no_parallel: Disable parallel simulation even if possible
            randomize_uxsim: Whether to run uxsim in random mode
            _unsafe_traffic_state_to_load: Alternative way to load UXSim states (unsafe, for advanced users only, at your own risk!)
            _unsafe_inst_state_to_load: Alternative way to load TrafficInst states (unsafe, for advanced users only, at your own risk!)
            _unsafe_plugin_state_to_load: Alternative way to load plugin states (unsafe, for advanced users only, at your own risk!)
        '''

        if plg_pool is None: plg_pool = PluginPool()
        if sta_pool is None: sta_pool = StaPool()

        self.__mpQ = mpQ
        self.__silent = silent
        self.__vb = vb
        self.__clntID = clntID
        if self.__mpQ:
            assert clntID != -1, Lang.ERROR_CLIENT_ID_NOT_SPECIFIED
            self.__silent = True
            self.__vb = None

        # Check if the folder exists
        proj_dir = Path(cfgdir)
        if not proj_dir.exists() or not proj_dir.is_dir():
            raise FileNotFoundError(f"Invalid project directory: {cfgdir}")
        
        # Determine result folder       
        pres = _calc_output_folder(cfgdir, outdir, outdir_direct)
        self.__outdir_direct = str(pres)
        self.__outdir = str(pres.parent)

        # Check if there is a previous results   
        if pres.is_dir() and (pres / TRIP_EVENT_LOG).exists():
            tm = time.strftime("%Y%m%d_%H%M%S", time.localtime(pres.stat().st_mtime))
            tm2 = 0
            while True:
                tm2 += 1
                new_path = f"{str(pres)}_{tm}_{tm2}"
                if not os.path.exists(new_path):
                    break
            if (pres / SAVED_STATE_FOLDER).exists() and load_last_state:
                initial_state = os.path.join(new_path, SAVED_STATE_FOLDER)
            pres.rename(new_path)
        pres.mkdir(parents=True, exist_ok=True)
        self.__pres = pres

        # Create simulation info log file
        self.__out = open(pres / SIM_INFO_LOG, "w", encoding="utf-8")

        # Record all __init__ parameters to file
        frame = inspect.currentframe()
        assert frame is not None
        args, _, _, values = inspect.getargvalues(frame)
        self.__out.write("Parameters:\n")
        for arg in args:
            if arg in ('self', 'vb', 'mpQ') or 'pool' in arg:
                continue
            self.__out.write(f"  {arg}: {values[arg]}\n")

        if gen_veh_command != "" or gen_scs_command != "" or gen_fcs_command != "":
            traff_gen = TrafficGenerator(str(proj_dir), silent)
            if gen_fcs_command != "":
                traff_gen.FCSFromArgs(gen_fcs_command)
                self.__print(Lang.INFO_REGEN_FCS)
                self.__mpsend("fcs:done")
            if gen_scs_command != "":
                traff_gen.SCSFromArgs(gen_scs_command)
                self.__print(Lang.INFO_REGEN_SCS)
                self.__mpsend("scs:done")
            if gen_veh_command != "":
                vehicles = traff_gen.EVTripsFromArgs(gen_veh_command)
                self.__print(Lang.INFO_REGEN_VEH)
                self.__mpsend("veh:done")
        else:
            vehicles = None
        
        proj_cfg = DetectFiles(str(proj_dir))

        if proj_cfg.py:
            with open(proj_cfg.py,"r",encoding="utf-8") as f:
                code = f.read()
                exec(code)
        
        # Create a simulation instance        
        if initial_state != "":
            self.__inst = TrafficInst.load(initial_state, Path(self.__outdir_direct) / TRIP_EVENT_LOG)
        elif _unsafe_inst_state_to_load is not None and _unsafe_traffic_state_to_load is not None:
            self.__inst = TrafficInst._partial_load_unsafe(_unsafe_inst_state_to_load, Path(self.__outdir_direct) / TRIP_EVENT_LOG)
            from .traffic.paraworlds import _load_world_unsafe
            w = _load_world_unsafe(_unsafe_traffic_state_to_load)
            assert isinstance(w, (SingleWorld, ParaWorlds))
            self.__inst.W = w
        
        if hasattr(self, "_V2SimInstance__inst"):
            self.__inst.silent = silent
            proj_cfg.net = self.__inst.net_file
            self.__print(Lang.INFO_NET.format(proj_cfg.net))
            self.__fcs_file = proj_cfg.fcs = self.__inst.fcs_file
            self.__print(Lang.INFO_FCS.format(proj_cfg.fcs, len(self.__inst.FCSList)))
            self.__scs_file = proj_cfg.scs = self.__inst.scs_file
            self.__print(Lang.INFO_SCS.format(proj_cfg.scs, len(self.__inst.SCSList)))
            self.__veh_file = proj_cfg.veh = self.__inst.veh_file
            self.__print(Lang.INFO_TRIPS.format(proj_cfg.veh, len(self.__inst.vehicles)))
            self.__start_time = self.__inst.start_time
            self.__actual_start_time = self.__inst.current_time
            self.__end_time = self.__inst.end_time
            self.__break_at = break_at if break_at >= 0 else self.__end_time
            self.__sim_dur = self.__break_at - self.__actual_start_time
            self.__steplen = self.__inst.step_len
            self.__print(Lang.INFO_TIME.format(self.__start_time, self.__end_time, self.__break_at, self.__steplen))
            self.routing_algo = self.__inst.routing_algo
            self.show_uxsim_info = self.__inst.show_uxsim_info
        else:
            # Detect road network file
            if not proj_cfg.net:
                raise RuntimeError(Lang.ERROR_NET_FILE_NOT_SPECIFIED)
            else:
                rnet_file = proj_cfg.net
            self.__print(Lang.INFO_NET.format(rnet_file))
            
            # Check vehicles and trips
            if not proj_cfg.veh:
                raise FileNotFoundError(Lang.ERROR_TRIPS_FILE_NOT_FOUND)
            self.__veh_file = proj_cfg.veh
            if vehicles is None:
                vehicles = EVDict(self.__veh_file)
            self.__print(Lang.INFO_TRIPS.format(self.__veh_file, len(vehicles)))

            # Check FCS file
            if not proj_cfg.fcs:
                raise FileNotFoundError(Lang.ERROR_FCS_FILE_NOT_FOUND)
            self.__fcs_file = proj_cfg.fcs
            fcs_obj:CSList[FCS] = CSList(filePath = self.__fcs_file, csType = FCS)
            self.__print(Lang.INFO_FCS.format(self.__fcs_file, len(fcs_obj)))

            # Check SCS file
            if not proj_cfg.scs:
                raise FileNotFoundError(Lang.ERROR_SCS_FILE_NOT_FOUND)
            self.__scs_file = proj_cfg.scs
            scs_obj:CSList[SCS] = CSList(filePath = self.__scs_file, csType = SCS)
            self.__print(Lang.INFO_SCS.format(self.__scs_file, len(scs_obj)))

            # Check start and end time
            if start_time == -1 or end_time == -1:
                raise ValueError(Lang.ERROR_ST_ED_TIME_NOT_SPECIFIED)
            self.__start_time = start_time
            self.__actual_start_time = self.__start_time
            self.__end_time = end_time
            self.__break_at = break_at if break_at >= 0 else self.__end_time
            self.__sim_dur = self.__break_at - self.__actual_start_time
            self.__steplen = traffic_step
            self.__print(Lang.INFO_TIME.format(start_time, end_time, self.__break_at, traffic_step))

            self.routing_algo = route_algo
            self.show_uxsim_info = show_uxsim_info

            self.__inst = TrafficInst(
                rnet_file, start_time, traffic_step, 
                end_time, str(pres / TRIP_EVENT_LOG), seed,
                vehfile = self.__veh_file, veh_obj = vehicles,
                fcsfile = self.__fcs_file, fcs_obj = fcs_obj,
                scsfile = self.__scs_file, scs_obj = scs_obj,
                routing_algo = route_algo,
                show_uxsim_info = show_uxsim_info,
                randomize_uxsim = randomize_uxsim,
                no_parallel = no_parallel,
                silent = silent or (self.__mpQ is not None),
            )

        # Enable plugins
        if initial_state != "":
            plugin_state_file = Path(initial_state) / PLUGINS_FILE
            with gzip.open(plugin_state_file, "rb") as f:
                d = pickle.load(f)
            _unsafe_plugin_state_to_load = d
                
        if _unsafe_plugin_state_to_load is not None:
            d = _unsafe_plugin_state_to_load
            assert isinstance(d, dict) and "obj" in d and "version" in d and "pickler" in d, Lang.INVALID_PLUGIN_STATES.format(plugin_state_file)
            plugin_state = d["obj"]
            assert isinstance(plugin_state, dict), Lang.INVALID_PLUGIN_STATES.format(plugin_state_file)
            assert CheckPyVersion(d["version"]), Lang.PY_VERSION_MISMATCH_PLG.format(d["version"], PyVersion())
            assert d["pickler"] == pickle.__name__, Lang.PICKLER_MISMATCH_PLG.format(d["pickler"], pickle.__name__)
        else:
            plugin_state = None

        if proj_cfg.plg:
            self.__plg_file = proj_cfg.plg
            disabled_plugins = list(map(lambda x: x.strip().lower(), no_plg.split(",")))
            self.__plgman = PluginMan(self.__plg_file, pres, self.__inst, disabled_plugins, plg_pool, plugin_state)
        else:
            self.__plgman = PluginMan(None, pres, self.__inst, [], plg_pool, plugin_state)
        
        # Find the power grid plugin
        self.__gridplg = None
        for plugname, plugin in self.__plgman.GetPlugins().items():
            if isinstance(plugin, PluginPDN):
                self.__gridplg = plugin
            self.__print(Lang.INFO_PLG.format(plugname, plugin.Description))

        # Create a data logger
        log_item = log.strip().lower().split(",")
        if len(log_item) == 1 and log_item[0] == "": log_item = []
        self.__sta = StaWriter(pres, self.__inst, self.__plgman.GetPlugins(), sta_pool, log_item)

        self.__copy = copy
        self.__plot_cmd = plot_command
        self.__proj_cfg = proj_cfg
        self.__proj_dir = proj_dir
        self.__working_flag = False
        self.save_on_abort = save_on_abort
        self.save_on_finish = save_on_finish
        self.copy_state = copy_state

        if alt_command is not None:
            assert initial_state == "", Lang.ALT_COMMAND_NOT_SUPPORTED
            for k, v in alt_command.items():
                if k == "start_time":
                    self.__start_time = int(v)
                elif k == "end_time":
                    self.__end_time = int(v)
                elif k == "traffic_step":
                    self.__steplen = int(v)
                elif k == "scs_slots":
                    for s in scs_obj:
                        s._slots = int(v)
                elif k == "fcs_slots":
                    for f in fcs_obj:
                        f._slots = int(v)

    @property
    def project_dir(self):
        '''Folder of the project'''
        return self.__proj_dir
    
    @property
    def result_dir(self):
        '''Folder of results'''
        return self.__outdir
    
    @property
    def result_dir_direct(self):
        '''Direct output folder'''
        return self.__outdir_direct
    
    @property
    def plot_command(self):
        '''Command for post-simulation plotting'''
        return self.__plot_cmd
    
    @property
    def ctime(self):
        '''Current simulation time, in second'''
        return self.__inst.current_time
    
    @property
    def step_length(self):
        '''Step length, in second'''
        return self.__steplen
    
    @property
    def btime(self):
        '''Simulation start time, in second'''
        return self.__start_time
    
    @property
    def etime(self):
        '''Simulation end time, in second'''
        return self.__end_time
    
    @property
    def copy(self):
        '''Indicate whether copy the source after simulation'''
        return self.__copy
    
    @property
    def clientID(self):
        '''Client ID in multiprocessing simulation'''
        return self.__clntID
    
    @property
    def silent(self):
        '''Indicate whether disable output'''
        return self.__silent
    
    @property
    def files(self):
        '''Files in the project'''
        return self.__proj_cfg
    
    @property
    def plugins(self):
        '''Plugins in the project'''
        return self.__plgman
    
    @property
    def statistics(self):
        '''Statistics in the project'''
        return self.__sta
    
    @property
    def core(self):
        '''Simulation core'''
        return self.__inst
    
    @property
    def fcs(self):
        '''List of FCSs'''
        return self.__inst.FCSList
    
    @property
    def scs(self):
        '''List of SCSs'''
        return self.__inst.SCSList
    
    @property
    def vehicles(self):
        '''Dict of vehicles'''
        return self.__inst.vehicles
    
    @property
    def edges(self):
        '''List of the edges'''
        return self.__inst.edges
    
    @property
    def edge_names(self):
        '''Name list of the edges'''
        return self.__inst.get_edge_names()
    
    @property
    def veh_count(self):
        '''Number of vehicles'''
        return len(self.__inst.vehicles)
    
    @property
    def is_working(self):
        '''Determine whether the simulation has started'''
        return self.__working_flag
    
    @property
    def pdn(self) -> Optional[PluginPDN]:
        '''Power grid plugin'''
        return self.__gridplg

    @property
    def trips_logger(self) -> TripsLogger:
        '''Trip logger'''
        return self.__inst.trips_logger
    
    @property
    def traffic_simulator(self):
        '''Traffic simulator object'''
        return self.__inst.W
    
    def send_to_host(self, command:str, obj:Any = None):
        '''Send message to host process'''
        assert self.__mpQ is not None, Lang.NO_HOST_EXISTS
        self.__mpsend(command, obj)
    
    def start(self):
        '''
        Start simulation.
            If you use this function, do not use function 'simulation'.
            Follow the start - step - stop paradigm.
        '''
        self.__working_flag = True
        self.__inst.simulation_start()
        self.__plgman.PreSimulationAll()
    
    def step(self) -> int:
        '''
        Simulation steps. 
            If you use this function, do not use function 'simulation'.
            Follow the start - step - stop paradigm.
        Return the simulation time after this step.
        '''
        t = self.__inst.current_time
        self.__plgman.PreStepAll(t)
        self.__inst.simulation_step(self.__steplen)
        self.__plgman.PostStepAll(t)
        self.__sta.Log(t)
        return self.__inst.current_time
    
    def step_until(self, t:int) -> int:
        '''
        Simulation steps till time t. 
            If you use this function, do not use function 'simulation'.
            Follow the start - step - stop paradigm.
        Return the simulation time after stepping.
        '''
        while self.__inst.current_time < t:
            self.step()
        return self.__inst.current_time
    
    def _save_obj(self):
        """Save plugin states. Advanced users only, at your own risk!"""
        return pickle.dumps({
            "obj": self.__plgman.SaveStates(),
            "version": PyVersion(),
            "pickler": pickle.__name__
        })
    
    def save(self, folder:Union[str, Path]):
        '''Save the current state of the simulation'''
        p = Path(folder) if isinstance(folder, str) else folder
        self.__inst.save(p)
        with gzip.open(p / PLUGINS_FILE, "wb") as f:
            pickle.dump({
                "obj": self.__plgman.SaveStates(),
                "version": PyVersion(),
                "pickler": pickle.__name__
            }, f)
    
    def stop(self, save_state_to:Union[str, Path] = ""):
        '''
        Stop simulation.
            If you use this function, do not use function 'simulation'.
            Follow the start - step - stop paradigm.
        '''
        if save_state_to != "":
            self.save(save_state_to)
            if self.copy_state:
                shutil.copytree(save_state_to, self.__proj_dir / SAVED_STATE_FOLDER, dirs_exist_ok=True)
                self.__print(Lang.SAVED_STATE_COPIED)
        self.__plgman.PostSimulationAll()
        self.__inst.simulation_stop()
        self.__sta.close()
        self.__out.close()
        if self.__copy:
            shutil.copy(self.__veh_file, self.__pres / Path(self.__veh_file).name)
            shutil.copy(self.__fcs_file, self.__pres / Path(self.__fcs_file).name)
            shutil.copy(self.__scs_file, self.__pres / Path(self.__scs_file).name)
            shutil.copy(self.__plg_file, self.__pres / Path(self.__plg_file).name)
            self.__print(Lang.CASE_FILE_COPIED)
        self.__working_flag = False
    
    def simulate(self):
        '''
        Main simulation function
            If you use this function, do not use start - step - stop paradigm
        Returns:
            (Whether the simulation ends normally, TrafficInst instance, StaWriter instance)
        '''
        self.__stopsig = False

        def eh(signum, frame):
            self.__print()
            self.__print(Lang.MAIN_SIGINT)
            self.__mpsend("exit")
            self.__stopsig = True
        
        if self.__vb is None and self.__clntID == -1:
            signal.signal(signal.SIGINT, eh)
        
        self.__st_time = time.time()
        self.__last_print_time = 0
        self.__last_mp_time = 0
        self.__mpsend("sim:start")
        self.start()

        while self.__inst.current_time < self.__break_at:
            self.step()
            if self.__stopsig:
                if self.save_on_abort:
                    p = self.__pres / SAVED_STATE_FOLDER
                    p.mkdir(parents=True, exist_ok=True)
                    self.save(p)
                break
            if not self.show_uxsim_info:
                self._istep()
        
        self.__print()
        dur = time.time() - self.__st_time
        self.__out.write(Lang.MAIN_SIM_DONE.format(time2str(dur)))
        self.__out.close()
        self.__print(Lang.MAIN_SIM_DONE.format(time2str(dur)))
        self.stop(str(self.__pres / SAVED_STATE_FOLDER) if self.save_on_finish else "")
        self.__mpsend("sim:done")
        if self.__plot_cmd != "" and not self.__stopsig:
            AdvancedPlot().configure(self.__plot_cmd)
        self.__mpsend("plot:done")
        return not self.__stopsig, self.__inst, self.__sta

    def _istep(self):
        ctime = time.time()
        if ctime - self.__last_print_time > 1 or self.__inst.current_time >= self.__break_at:
            # Progress in command line updates once per second
            progress = 100 * (self.__inst.current_time - self.__actual_start_time) / self.__sim_dur
            eta = (
                time2str((ctime - self.__st_time) * (100 - progress) / progress)
                if ctime - self.__st_time > 3
                else "N/A"
            )
            self.__print("\r",end="")
            self.__print(
                Lang.MAIN_SIM_PROG.format(
                    round(progress,2), 
                    self.__inst.current_time, 
                    self.__break_at, 
                    time2str(ctime-self.__st_time), 
                    eta
                ),
                end="",
            )
            if ctime - self.__last_mp_time > 5:
                # Communicate with the main process every 5 seconds in multi-process mode
                self.__mpsend(f"sim:{progress:.2f}")
                self.__last_mp_time = ctime
            self.__last_print_time = ctime
    
    def __del__(self):
        if hasattr(self, "_V2SimInstance__out") and hasattr(self.__out, "closed") and not self.__out.closed:
            self.__out.close()

def simulate_single(vb=None, **kwargs)->bool:
    '''
    Single process simulation
        vb: Visualization window. None means no visualization.
        kwargs: Simulation parameters. Use function 'get_sim_params' to get.
    '''
    return V2SimInstance(**kwargs, vb=vb, silent=False).simulate()[0]

def simulate_multi(mpQ:Optional[queue.Queue], clntID:int, **kwargs)->bool:
    '''
    Multi-process simulation
        mpQ: Queue for communication with the main process.
        clntID: Client ID.
        kwargs: Simulation parameters. Use function 'get_sim_params' to get.
    '''
    return V2SimInstance(**kwargs, mpQ=mpQ, clntID=clntID, silent=True).simulate()[0]

if __name__ == "__main__":
    print(Lang.CORE_NO_RUN)


"""
General utilities for UXsim.
"""

import warnings
import functools
import traceback
import sys

import numpy as np
import matplotlib.pyplot as plt


def catch_exceptions_and_warn(warning_msg=""):
    """
    A decorator that catches exceptions in the decorated function and raises a warning with the specified message.

    Parameters
    ----------
    warning_msg : str, optional
        A custom message to be displayed along with the exception details. Default is an empty string.

    Returns
    -------
    decorator
        A decorator that wraps the decorated function to catch exceptions and raise warnings.

    Notes
    -----
    When an exception occurs in the decorated function, the decorator will raise a warning that includes:
    - The name of the function.
    - A custom message, if provided.
    - The type of the exception.
    - The exception message.
    - The filename and line number where the decorated function was called.

    This decorator is for inessential functions where random exceptions are expected (especially due to file I/O or something), but you want to avoid crashing the entire program and instead receive a warning. Mainly written by ChatGPT.
    """
    warnings.simplefilter('default', category=UserWarning)
    warnings.formatwarning = lambda message, category, filename, lineno, line: f"{category.__name__}: {message}\n"
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                #warnings.warn(f"{func.__name__}(){warning_msg}: {type(e).__name__}: {str(e)}")
                current_frame = sys._getframe()
                caller_frame = traceback.extract_stack(current_frame, limit=2)[0]
                filename = caller_frame.filename
                lineno = caller_frame.lineno
                warnings.warn(f"{filename}:{lineno}:{func.__name__}(){warning_msg}: {type(e).__name__}: {str(e)}")
        return wrapper
    return decorator


def lange(l):
    """
    Super lazy function for abbreviating range(len(l)).

    Parameters
    ----
    l : list
    """
    return range(len(l))

def printtry(*args, **kwargs):
    """
    Print messages safely, evaluating any callable arguments and catching exceptions. Mainly written by GPT-4.

    Parameters
    ----------
    *args : any
        The contents to print wrapped by lambda. E.g., lambda: (x, y, z)
    **kwargs : any
        Arbitrary keyword arguments. These are passed directly to the built-in print function.

    Examples
    --------
    >>> printtry(lambda: (mylist[100], mylist[1000], mylist[10000]))
    >>> printtry(lambda: (mylist[100], mylist[1000], mylist[10000]), sep=";", end="!")
    """
    try:
        # Evaluate each argument and print them
        evaluated_args = [arg() if callable(arg) else arg for arg in args][0]
        print(*evaluated_args, **kwargs)
    except Exception as e:
        # Print the exception if it occurs
        print("EXCEPTION:", e)
    

def get_font_for_matplotlib(fname=None):
    """
    Get a multi-language (currently Japanese only) font for matplotlib.
    TODO: check if it works on different OS. It is only checked on Japanese Windows. 
    TODO: explore if it can be extended for other languages.

    Returns
    -------
    str
        The name of a Japanese font that is available on the system. If no Japanese font is found, "monospace" is returned.
    """
    from matplotlib import font_manager

    font_list = font_manager.findSystemFonts()

    if fname is not None:
        font = fname
    elif any("Noto Sans Mono CJK JP" in font.lower() for font in font_list):
        font = "Noto Sans Mono CJK JP"
    elif any("msgothic" in font.lower() for font in font_list):
        font = "MS Gothic"
    else:
        font = "monospace"
        
    return font

def print_columns(*lists):
    """
    Convinient function to check contents of 1d lists. For debug.
    """
    # Determine the maximum length of the lists
    max_length = max(len(lst) for lst in lists)

    # Iterate through the maximum number of rows
    for i in range(max_length):
        # For each list, print the element at the current row or space if out of range
        for lst in lists:
            if i < len(lst):
                try:
                    print(f"{lst[i]:<10}", end=" ")  # Adjust the width as necessary
                except TypeError:
                    print(f"{str(lst[i]):<10}", end=" ")
            else:
                print(" " * 10, end=" ")  # Adjust spacing to match the above width
        print()  # Newline after each row

class LoggingWarning(UserWarning):
    """
    This warns that when vehicle_logging_timestep_interval is not 1 but called vehicle logging-related functions.
    """
    pass

def plot_multiple_y(x=None, ys=None, labels=None, **kwarg):
    """
    Plot multiple y values on the same plot, normalizing each value to [0, 1].

    Parameters
    ----------
    x : array-like, optional
        The x values to plot. If None, the indices of the y values are used. Default is None.
    ys : list of array-like
        The y values to plot.
    labels : list of str, optional
        The labels for each y value.
    **kwarg : any
        Additional keyword arguments to pass to the plt.plot function
    """
    if labels is None:
        labels = [f"data-{i}" for i in lange(ys)]

    for i,y in enumerate(ys):
        y_min = np.min(y)
        y_max = np.max(y)
        if x is not None:
            plt.plot(x, (y-y_min)/(y_max-y_min), label=f"{labels[i]}: [{y_min:.5g}, {y_max:.5g}]", **kwarg)
        else:
            plt.plot((y-y_min)/(y_max-y_min), label=f"{labels[i]}: [{y_min:.5g}, {y_max:.5g}]", **kwarg)

    plt.ylabel("normalized to [0, 1]")
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))


def equal_tolerance(val, check, rel_tol=0.1, abs_tol=0.0):
    """
    Function for tests.
    """
    if check == 0 and abs_tol == 0:
        abs_tol = 0.1
    return abs(val - check) <= abs(check*rel_tol) + abs_tol


from itertools import chain
from typing import Optional
from feasytools import TimeFunc, LangLib
from .base import *

FILE_GEN = "gen"
FILE_BUS = "bus"
FILE_LINE = "line"
FILE_PVW = "pvw"
FILE_ESS = "ess"

GEN_ATTRIB = ["P","Q","costp"]
GEN_TOT_ATTRIB = ["totP","totQ","totC"]
BUS_ATTRIB = ["Pd","Qd","Pg","Qg","V"]
BUS_TOT_ATTRIB = ["totPd","totQd","totPg","totQg"]
LINE_ATTRIB = ["P","Q","I"]
PVW_ATTRIB = ["P","curt"]
ESS_ATTRIB = ["P","soc"]

_L = LangLib(["en", "zh_CN"])
_L.SetLangLib("en",
    GEN = "Generator",
    BUS = "Bus",
    LINE = "Line",
    PVW = "PV/Wind",
    ESS = "ESS",
)

_L.SetLangLib("zh_CN",
    GEN = "发电机",
    BUS = "母线",
    LINE = "线路",
    PVW = "光伏/风电",
    ESS = "储能",
)

def _chk(x:Optional[float])->float:
    if x is None: return 0
    return x

def _find_grid_plugin(plugins:Dict[str,PluginBase])->IGridPlugin:
    for plg in plugins.values():
        if isinstance(plg, IGridPlugin):
            return plg
    raise ValueError("No plugin for grid found.")

class StaGen(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        self.__plg = _find_grid_plugin(plugins)
        gen_names = self.__plg.Grid.GenNames
        super().__init__(FILE_GEN,path,cross_list(gen_names,GEN_ATTRIB)+GEN_TOT_ATTRIB,tinst,plugins)
    
    @staticmethod
    def GetLocalizedName() -> str:
        return _L("GEN")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["pdn"]

    def GetData(self,inst:TrafficInst,plugins:List[PluginBase])->Iterable[Any]:
        mpdn = self.__plg
        sb_MVA = mpdn.Grid.Sb_MVA
        _t = inst.current_time
        p = []; q = []; cp = []
        for g in mpdn.Grid.Gens:
            costthis = g.Cost(_t)
            if costthis is None:
                costthis = 0
            if g.P is None or g.Q is None:
                p.append(0); q.append(0)
                cp.append(costthis)
            else:
                p.append(g.P*sb_MVA); q.append(g.Q*sb_MVA)
                cp.append(costthis)
        return chain(p,q,cp,[sum(p), sum(q), sum(cp)])

class StaBus(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        self.__plg = _find_grid_plugin(plugins)
        bus_names = self.__plg.Grid.BusNames
        self.__bus_with_gens = [b.ID for b in self.__plg.Grid.Buses if len(self.__plg.Grid.GensAtBus(b.ID))>0]
        super().__init__(FILE_BUS,path,cross_list(bus_names,["Pd","Qd","V"]) 
            + cross_list(self.__bus_with_gens,["Pg","Qg"]) + BUS_TOT_ATTRIB,tinst,plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("BUS")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["pdn"]
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        mpdn = self.__plg.Grid
        sb_MVA = mpdn.Sb
        _t = inst.current_time
        bs = mpdn.Buses
        Pd = [b.Pd(_t)*sb_MVA for b in bs]
        Qd = [b.Qd(_t)*sb_MVA for b in bs]
        V = (b.V*mpdn.Ub if b.V else 0 for b in bs)
        Pg = []; Qg = []
        for bn in self.__bus_with_gens:
            pg = 0; qg = 0
            for g in mpdn.GensAtBus(bn):
                if isinstance(g.P, float): pg += g.P
                elif isinstance(g.P, TimeFunc): pg += g.P(_t)
                if isinstance(g.Q, float): pg += g.Q
                elif isinstance(g.Q, TimeFunc): pg += g.Q(_t)
            Pg.append(pg*sb_MVA); Qg.append(qg*sb_MVA)
        return chain(Pd,Qd,V,Pg,Qg,[sum(Pd), sum(Qd), sum(Pg), sum(Qg)]) # Unit = MVA

class StaLine(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        self.__plg = _find_grid_plugin(plugins)
        super().__init__(FILE_LINE,path,cross_list(self.__plg.Grid._lines.keys(),LINE_ATTRIB),tinst,plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("LINE")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["pdn"]
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        mpdn = self.__plg.Grid
        P = (_chk(b.P)*mpdn.Sb for b in mpdn.Lines)
        Q = (_chk(b.Q)*mpdn.Sb for b in mpdn.Lines)
        I = (_chk(b.I)*mpdn.Ib for b in mpdn.Lines)
        return chain(P,Q,I) # Unit = MVA or kA

class StaPVWind(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        self.__plg = _find_grid_plugin(plugins)
        super().__init__(FILE_PVW, path, cross_list(self.__plg.Grid._pvws.keys(),PVW_ATTRIB),tinst,plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("PVW")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["pdn"]
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        mpdn = self.__plg.Grid
        P = (b.P(inst.current_time)*mpdn.Sb for b in mpdn.PVWinds)
        curt = (_chk(b._cr) for b in mpdn.PVWinds)
        return chain(P, curt) # Unit = MVA or %

class StaESS(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        self.__plg = _find_grid_plugin(plugins)
        super().__init__(FILE_ESS, path, cross_list(self.__plg.Grid._esss.keys(),ESS_ATTRIB),tinst,plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("ESS")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["pdn"]
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        mpdn = self.__plg.Grid
        P = (_chk(b.P)*mpdn.Sb for b in mpdn.ESSs)
        soc = (b.SOC for b in mpdn.ESSs)
        return chain(P, soc) # Unit = MVA or %